#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'eugene'

"""

FANTASTIC EXAMPLES OF UML -> C++ here :

http://icarus.cs.weber.edu/~dab/cs1410/textbook/chapters.html
->
http://icarus.cs.weber.edu/~dab/cs1410/textbook/uml.html

TODO : Parameterized constructors and inheritance!

"""
try:
    import vppfs
except (ModuleNotFoundError, ImportError) as e:
    from kojen import vppfs

import unittest

'''----------------------------------------------------------------------'''
''' These appear to be the values for the different types of visibility '''


class Visibility:
    Public = '71'
    Protected = '67'
    Private = '66'
    Package = '68'


def VisibilityToHumanReadableString(vis):
    if vis == Visibility.Public:
        return "public"
    if vis == Visibility.Protected:
        return "protected"
    if vis == Visibility.Private:
        return "private"
    if vis == Visibility.Package:
        return "package"
    return "public"  # It appears that "unspecified" can return '65'


'''----------------------------------------------------------------------'''
''' These appear to be the values for the different types of scope (static = classifier)'''
SCOPE_CLASSIFIER = '65'

'''----------------------------------------------------------------------'''
'''A parameter of a function, with direction 'in' means it can't be modified...thus const '''
DIRECTION_IN = '65'
DIRECTION_OUT = '66'

'''----------------------------------------------------------------------'''
'''An association has 3 types ... association, aggregation, composition '''

# Default is association...each know about the other...
AGGREGATION_KIND__AGGREGATE  =  '66'
AGGREGATION_KIND__COMPOSITE  =  '67'

'''----------------------------------------------------------------------'''


def CleanName(name):
    return name.replace('=', '_').replace('<', '_').replace('>', '_').replace(';', '_').replace(r'\n', '_').\
        replace(r'\r', '_').replace(r'\t', '_').replace('\n', '_').replace('\r', '_').replace('\t', '_').\
        replace('"', "_").replace('(', '_').replace(')', '_').replace("*", "_").replace("+", "_").replace("-", "_").\
        replace("^", "_").replace("%", "_").replace("~", "_").replace("!", "_").replace(" ", "").replace("|", "_").\
        replace(":", "_")


def GetNestedTypeNamesFromNestedTypeIDS(nestedTypeIDs, table_vppmodelelements):
    # If the type is a class in a package, type will be 'packageID:classID' etc etc  for levels of nesting...
    # For most OOP languages we need to nest the type...
    nested_types = nestedTypeIDs.split(':')
    result = ""
    for t in nested_types:
        modelelement = table_vppmodelelements.GetModelElement(t)
        result = result + modelelement.NAME + "::"
    # remove the last "::"
    return result.rstrip(':')

PRIMITIVES = {"char",
              "wchar_t",
              "string",
              "wstring",
              "std::string",
              "std::wstring",
              "signed char",
              "short int",
              "int",
              "long int",
              "unsigned char",
              "unsigned short int",
              "unsigned int",
              "unsigned long int",
              "wchar_t",
              "bool",
              "boolean",
              "float",
              "double",
              "long",
              "void",
              "size_t",
              "ptrdiff_t",
              "int8",
              "uint8",
              "int16",
              "uint16",
              "int32",
              "uint32",
              "int64",
              "uint64",
              "int8_t",
              "uint8_t",
              "int16_t",
              "uint16_t",
              "int32_t",
              "uint32_t",
              "int64_t",
              "uint64_t"}

def IsTypePrimitive(type):
    global PRIMITIVES
    type = CleanModifiersFromType(type)
    return type in PRIMITIVES

def IsTypePointerOrRef(type):
    return type.find("*") > -1 or type.find("&") > -1

def CleanModifiersFromType(type):
    # &, *, []
    return type.replace("*", "").replace("&", "").replace("]","").replace("[","").replace("boolean", "bool")

'''----------------------------------------------------------------------'''
''' class Inheritance -> from searching the web, it appears the UML 'Generalization' and 'Realization' are the same, for C++ at least.
    This, until I find a better source of info (or a language with different semantics), I will treat them both the same.
    
    class Generalization(vppfs.VPPModelElement):
    class Realization(vppfs.VPPModelElement):
    
    If we were to split these concepts, the same code would be used.
'''

class Inheritance(vppfs.VPPModelElement):
    """ This class represents a 'Association' VPPModelElement.
        In UML, this is just a namespace name.
        It will contain the IDs of the classes that are contained within.
    """

    def __init__(self, baseclass, table_vppmodelelements, is_realization):
        self.ID = baseclass.ID
        self.MODEL_TYPE = baseclass.MODEL_TYPE
        self.PARENT_ID = baseclass.PARENT_ID
        self.NAME = baseclass.NAME
        self.BLOB_STRING = baseclass.BLOB_STRING
        self.table_vppmodelelements = table_vppmodelelements
        self.dict_from_BLOB_STRING = {}
        self.IS_REALIZATION = is_realization

        self.CLASS_FROM = ""  # For some reason these are not fully qualified. So this is just for debug.
        self.CLASS_FROM_ID = ""  # Perhaps the class needs to be gotten when generating code that uses this...
        self.CLASS_TO = ""  # For some reason these are not fully qualified. So this is just for debug.
        self.CLASS_TO_ID = ""  # Perhaps the class needs to be gotten when generating code that uses this...

        self.Parse()

    def Parse(self):
        self.dict_from_BLOB_STRING = vppfs.ParseBLOB_Recursive(self.BLOB_STRING)
        #vppfs.Print_RecursiveParsed_BLOB(self.dict_from_BLOB_STRING)

        for k, v in self.dict_from_BLOB_STRING.items():
            if k.lower().find('child') > -1:
                self.CLASS_FROM = GetNestedTypeNamesFromNestedTypeIDS(v['fromModel_0'],self.table_vppmodelelements)
                self.CLASS_FROM_ID = v['fromModel_0'].split(':')[-1]  # -1 = last element...gotto love python
                self.CLASS_TO = GetNestedTypeNamesFromNestedTypeIDS(v['toModel_0'],self.table_vppmodelelements)
                self.CLASS_TO_ID = v['toModel_0'].split(':')[-1]  # -1 = last element...gotto love python

    def PostProjectParseFix(self, classDiagram):
        try:
            cl_fr = classDiagram.classes[self.CLASS_FROM_ID]
            self.CLASS_FROM = cl_fr.NAMESPACE + "::" + cl_fr.NAME
        except KeyError:
            print("Warning : class " + self.CLASS_FROM + " does not exist.")

        try:
            cl_to = classDiagram.classes[self.CLASS_TO_ID]
            self.CLASS_TO = cl_to.NAMESPACE + "::" + cl_to.NAME
        except KeyError:
            print("Warning : class " + self.CLASS_TO + " does not exist.")



'''----------------------------------------------------------------------'''


class Association(vppfs.VPPModelElement):
    """ This class represents a 'Association' VPPModelElement.
        In UML, this is just a namespace name.
        It will contain the IDs of the classes that are contained within.
    """

    '''
    <ASSOCIATION> 
    http://icarus.cs.weber.edu/~dab/cs1410/textbook/11.Relationships/association.html
    class B{
        A* a;
    };
    class A{
        B* b;
    };
    
    <AGGREGATION>
    http://icarus.cs.weber.edu/~dab/cs1410/textbook/11.Relationships/aggregation.html
    class B {};
    class A
    {
        B* b;
    };
    
    <COMPOSITION>
    class B {};
    class A
    {
        B b;
    };
    '''

    def __init__(self, baseclass, table_vppmodelelements):
        self.ID = baseclass.ID
        self.MODEL_TYPE = baseclass.MODEL_TYPE
        self.PARENT_ID = baseclass.PARENT_ID
        self.NAME = baseclass.NAME
        self.BLOB_STRING = baseclass.BLOB_STRING
        self.table_vppmodelelements = table_vppmodelelements
        self.dict_from_BLOB_STRING = {}
        self.TYPE = "Association"
        self.USER_COMMENTS = ""
        self.CLASS_FROM = ""
        self.CLASS_FROM_ID = ""  # Perhaps the class needs to be gotten when generating code that uses this...
        self.CLASS_FROM_VISIBILITY = "private"
        self.CLASS_FROM_IS_STATIC = False
        self.CLASS_FROM_IS_CONST = False
        self.CLASS_FROM_MULTIPLICITY = "0..1" # 'Class From' multiplicity of 0 does not make sense with composition.
        self.CLASS_FROM_HAS_GETTER = False
        self.CLASS_FROM_HAS_SETTER = False
        self.CLASS_TO = ""
        self.CLASS_TO_ID = ""  # Perhaps the class needs to be gotten when generating code that uses this...
        self.CLASS_TO_VISIBILITY = "private"
        self.CLASS_TO_IS_STATIC = False
        self.CLASS_TO_IS_CONST = False
        self.CLASS_TO_MULTIPLICITY = "0..1"  # This only makes sense in an association where both know about each other...
        self.CLASS_TO_HAS_GETTER = False
        self.CLASS_TO_HAS_SETTER = False
        self.Parse()

    def Parse(self):
        self.dict_from_BLOB_STRING = vppfs.ParseBLOB_Recursive(self.BLOB_STRING)
        #vppfs.Print_RecursiveParsed_BLOB(self.dict_from_BLOB_STRING)
        self.ParseAssociation()

    def ParseAssociation(self):
        for k, v in self.dict_from_BLOB_STRING.items():
            if 'documentation_plain' in v:
                self.USER_COMMENTS = v['documentation_plain']
            if k.lower().find('child') > -1:
                for kk, vv in v.items():
                    if kk.lower().find('child') > -1:
                        if vv:
                            if vv['type'].lower().find('associationend') != -1:
                                for kkk, vvv in vv.items():
                                    if kkk.lower().find('child') > -1:
                                        # Direction ...
                                        if vvv['Direction'] == '0':
                                            self.CLASS_FROM = GetNestedTypeNamesFromNestedTypeIDS(vvv['EndModelElement_0'], self.table_vppmodelelements)
                                            self.CLASS_FROM_ID = vvv['EndModelElement_0'].split(':')[-1]  # -1 = last element...gotto love python
                                        elif vvv['Direction'] == '1':
                                            self.CLASS_TO = GetNestedTypeNamesFromNestedTypeIDS(vvv['EndModelElement_0'], self.table_vppmodelelements)
                                            self.CLASS_TO_ID = vvv['EndModelElement_0'].split(':')[-1]  # -1 = last element...gotto love python
                                        # TYPE
                                        if 'aggregationKind' in vvv:
                                            if vvv['aggregationKind'] == AGGREGATION_KIND__AGGREGATE:
                                                self.TYPE = "Aggregation"
                                            elif vvv['aggregationKind'] == AGGREGATION_KIND__COMPOSITE:
                                                self.TYPE = "Composition"
                                        # Multiplicity ...
                                        if 'multiplicity' in vvv:
                                            m = vvv['multiplicity']
                                            if vvv['Direction'] == '0':
                                                self.CLASS_FROM_MULTIPLICITY = m
                                            elif vvv['Direction'] == '1':
                                                self.CLASS_TO_MULTIPLICITY = m
                                        else:
                                            # Set defaults depending on aggregation.
                                            if vvv['Direction'] == '0':
                                                if self.TYPE == "Composition":  # composition means, the class 'from' has an value (not a pointer) of the class 'to' ... thus can not be '0'
                                                    self.CLASS_FROM_MULTIPLICITY = "1"
                                            elif vvv['Direction'] == '1':
                                                if self.TYPE != "Association":  # If it is NOT association, then the class 'to' knows nothing about the class from (its a one way relationship).
                                                    self.CLASS_TO_MULTIPLICITY = '0'
                                        # Visibility
                                        # 'package' to me makes most sense as public, and static.
                                        # However, it appears that 'private' is the default...and thus signified by 'visibility' not being set here...
                                        if 'visibility' in vvv:
                                            vis = vvv['visibility']
                                            if vis == Visibility.Package:
                                                if vvv['Direction'] == '0':
                                                    self.CLASS_FROM_IS_STATIC = True
                                                elif vvv['Direction'] == '1':
                                                    self.CLASS_TO_IS_STATIC = True
                                            else:
                                                if vvv['Direction'] == '0':
                                                    self.CLASS_FROM_VISIBILITY = VisibilityToHumanReadableString(vis)
                                                elif vvv['Direction'] == '1':
                                                    self.CLASS_TO_VISIBILITY = VisibilityToHumanReadableString(vis)
                                        # Getter/Setter
                                        if 'providePropertyGetterMethod' in vvv:
                                            if vvv['Direction'] == '0':
                                                self.CLASS_FROM_HAS_GETTER = True
                                            elif vvv['Direction'] == '1':
                                                self.CLASS_TO_HAS_GETTER = True
                                        if 'providePropertySetterMethod' in vvv:
                                            if vvv['Direction'] == '0':
                                                self.CLASS_FROM_HAS_SETTER = True
                                            elif vvv['Direction'] == '1':
                                                self.CLASS_TO_HAS_SETTER = True
                                        # Readonly (const)
                                    if 'readOnly' in vvv:
                                        if vvv['Direction'] == '0':
                                            self.CLASS_FROM_IS_CONST = True
                                        elif vvv['Direction'] == '1':
                                            self.CLASS_TO_IS_CONST = True


'''----------------------------------------------------------------------'''


class Package(vppfs.VPPModelElement):
    """ This class represents a 'Package' VPPModelElement.
        In UML, this is just a namespace name.
        It will contain the IDs of the classes that are contained within.
    """
    def __init__(self, baseclass):
        self.ID = baseclass.ID
        self.MODEL_TYPE = baseclass.MODEL_TYPE
        self.PARENT_ID = baseclass.PARENT_ID
        self.NAME = baseclass.NAME
        self.BLOB_STRING = baseclass.BLOB_STRING
        # from UML Package (as a package can be nested within a package)...
        self.NAMESPACE = ""
        self.dict_from_BLOB_STRING = {}
        self.CLASSES_IN_PACKAGE = []
        self.Parse()

    def Parse(self):
        self.dict_from_BLOB_STRING = vppfs.ParseBLOB_Recursive(self.BLOB_STRING)
        self.ParseClassesInPackage()
        #Print_RecursiveParsed_BLOB(self.dict_from_BLOB_STRING)
        #print(self.BLOB_STRING)

    def ParseClassesInPackage(self):
        """ Returns the IDs of all the classes contained within (in a list)
        """
        for k, v in self.dict_from_BLOB_STRING.items():
            if k.lower().find('child') > -1:
                for kk, vv in v.items():
                    if kk.lower().find('child') > -1:
                        if isinstance(vv, str):  # there could be a child that is a dict
                            # Child_0 = pqLlew6GAqAA7wvG:4bxByw6GAqAA7wcw -> where THIS_ID:CLASS_ID
                            # Keep the fully qualified class here...it will be all nesting...so the namespace can be worked out at the end...
                            self.CLASSES_IN_PACKAGE.append(vv.strip())


'''----------------------------------------------------------------------'''


class ClassAttribute:
    def __init__(self, container, table_vppmodelelements):
        # Attribute Name
        self.NAME = ""
        if container:
            self.NAME = container['name']
        self.VISIBILITY = 'private'  # appears to be the default. So, when set to 'private' there is no attribute...
        self.TYPE_MODIFIER = ""
        self.USER_COMMENTS = ""
        self.TYPE = "void"
        self.MULTIPLICITY = ""
        self.UNIQUE = False
        self.HAS_SETTER = False
        self.HAS_GETTER = False
        self.IS_STATIC = False
        self.IS_CONST = False
        self.INITIAL_VALUE = None

        if not container or not table_vppmodelelements:
            return

        child_0 = container['child_0']
        # Attribute Visibility
        if 'visibility' in child_0:
            self.VISIBILITY = VisibilityToHumanReadableString(child_0['visibility'])
        # Attribute Type Modifier
        if 'typeModifier' in child_0:
            self.TYPE_MODIFIER = child_0['typeModifier']
        # Attribute Type ... dont allow users to put eg &,*,[] here...these should be in the modifier...force them
        if 'type_0' in child_0:
            self.TYPE = CleanModifiersFromType(GetNestedTypeNamesFromNestedTypeIDS(child_0['type_0'], table_vppmodelelements))
        # User documentation about this function
        if 'documentation_plain' in child_0:
            self.USER_COMMENTS = child_0['documentation_plain']
        # User requested setter
        if 'hasSetter' in child_0:
            self.HAS_SETTER = True
        # User requested getter
        if 'hasGetter' in child_0:
            self.HAS_GETTER = True
        # Static means the scope is 'classifier', meaning for all of the classifier of the class
        if 'scope' in child_0:
            if child_0['scope'] == SCOPE_CLASSIFIER:
                self.IS_STATIC = True
        if 'initialValue_string' in child_0:
            self.INITIAL_VALUE = child_0['initialValue_string']
        if 'readOnly' in child_0:
            self.IS_CONST = True
            # however, a const attribute MUST have an initial value...
            #if self.INITIAL_VALUE is None:
            #	raise Exception("'Read-only (const)' marked class attribute '" + self.NAME + "' requires an initial value to be set.")
            # Actually...rather create a special constructor for this...fuck.
        if 'multiplicity' in child_0:
            self.MULTIPLICITY = child_0['multiplicity']

    def From(self, NAME, TYPE, TYPE_ID, MODIFIER, MULTIPLICITY, IS_STATIC, IS_CONST, VISIBILITY, HAS_GETTER, HAS_SETTER, USER_COMMENTS):
        self.NAME = NAME
        self.ID = TYPE_ID
        self.TYPE_MODIFIER = MODIFIER
        self.USER_COMMENTS = ""
        self.TYPE = TYPE
        self.VISIBILITY = VISIBILITY
        self.MULTIPLICITY = MULTIPLICITY
        self.USER_COMMENTS = USER_COMMENTS
        self.UNIQUE = False
        self.HAS_SETTER = HAS_SETTER
        self.HAS_GETTER = HAS_GETTER
        self.IS_STATIC = IS_STATIC
        self.IS_CONST = IS_CONST
        self.INITIAL_VALUE = None


class ClassOperation:
    def __init__(self, container, table_vppmodelelements):
        # Operation Name
        self.NAME = container['name']
        self.VISIBILITY = 'public'
        self.RETURN_TYPE = "void"
        self.RETURN_TYPE_MODIFIER = ""
        self.PARAMETERS = []
        self.USER_COMMENTS = ""
        self.VIRTUAL = False
        self.IS_STATIC = False
        self.IS_CONST = False

        child_0 = container['child_0']
        # Operation Visibility
        if 'visibility' in child_0:
            self.VISIBILITY = VisibilityToHumanReadableString(child_0['visibility'])
            # Second static option? 'visibility' = package...make it public and static.
            if self.VISIBILITY.lower().strip() == "package":
                self.IS_STATIC = True
                self.VISIBILITY = "public"

        # Operation Return Type -> if someone does not set this...then assume it is void.
        if 'returnType_0' in child_0:  # ... dont allow users to put eg &,*,[] here...these should be in the modifier...force them
            self.RETURN_TYPE = CleanModifiersFromType(GetNestedTypeNamesFromNestedTypeIDS(child_0['returnType_0'], table_vppmodelelements))
        if 'typeModifier' in child_0:
            self.RETURN_TYPE_MODIFIER = child_0['typeModifier']
        # Parameters
        for k, v in child_0.items():
            if k.find("child") > -1:
                if v:  # test for empty dictionary (empty dictionaries evaluate to False)
                    if 'parameter' == v['type'].lower():
                        v_child_0 = v['child_0']
                        # Options here : a parameter is a either a basic type (int, double), or a model element (namespace::class)...
                        type_s = ""  # ... dont allow users to put eg &,*,[] here...these should be in the modifier...force them
                        if 'type_string' in v_child_0:  # basic type
                            type_s = CleanModifiersFromType(v_child_0['type_string'])
                        else:
                            type_s = CleanModifiersFromType(GetNestedTypeNamesFromNestedTypeIDS(v_child_0['type_0'], table_vppmodelelements))
                        # IN parameters cant be changed...const...
                        const = ""
                        direction = "inout"
                        if 'direction' in v_child_0:
                            if v_child_0['direction'] == DIRECTION_IN:  # An 'in' parameter, even if its a pointer, is CONST...for C++. For other languages...different! so pass the direction...
                                const = "const"
                                direction = "in"
                            if v_child_0['direction'] == DIRECTION_OUT:
                                direction = "out"
                        # undefined type modifier means nothing...
                        typemodifier = ""
                        if 'typeModifier' in v_child_0:
                            typemodifier = v_child_0['typeModifier']
                        # default value
                        defaultvalue = ""
                        if 'defaultValue_string' in v_child_0:
                            defaultvalue = v_child_0['defaultValue_string']
                        multiplicity = ""
                        if 'multiplicity' in v_child_0:
                            multiplicity = v_child_0["multiplicity"]
                        param = {'const': const, 'type': type_s, 'name': v['name'], 'modifier': typemodifier, 'defaultvalue': defaultvalue, 'multiplicity': multiplicity, 'direction':direction}
                        self.PARAMETERS.append(param)
        # User documentation about this function
        if 'documentation_plain' in child_0:
            self.USER_COMMENTS = child_0['documentation_plain']
        # Is this member abstract
        if 'abstract' in child_0:
            self.VIRTUAL = True
        ''' IS this member const? Declaring a member function with the const keyword specifies that the function is 
            a "read-only" function that does not modify the object for which it is called. 
            If func does not change anything on it ( it can change on Logger) It can be "labeled" as "query".
        '''
        if 'query' in child_0:
            self.IS_CONST = True
        # Static means the scope is 'classifier', meaning for all of the classifier of the class
        if 'scope' in child_0:
            if child_0['scope'] == SCOPE_CLASSIFIER:
                self.IS_STATIC = True


class Class(vppfs.VPPModelElement):
    """ This class represents a 'Class' VPPModelElement.
        All VPPModelElement comments hold true. This class is specialized
        with more fields : 'from state', 'to state', 'guard' and 'activity'.
        These added fields are the model element id's of the corresponding model elements.
        These are added via trickery in the 'Parse' function.
    """

    def __init__(self, baseclass, parent_classDiagram):
        self.parent_classDiagram = parent_classDiagram
        self.ID = baseclass.ID
        self.MODEL_TYPE = baseclass.MODEL_TYPE
        self.PARENT_ID = baseclass.PARENT_ID
        self.NAME = baseclass.NAME
        self.BLOB_STRING = baseclass.BLOB_STRING
        self.dict_from_BLOB_STRING = {}
        # From stereotypes/flags
        self.PURE_VIRTUAL_INTERFACE = False
        self.AUTOGEN = False
        self.USER_COMMENTS = ""
        # from UML Package
        self.NAMESPACE = ""
        # from blob
        self.OPERATIONS = []
        self.ATTRIBUTES = []
        # Enum support
        self.IS_ENUM = False
        self.ENUM_LITERALS = []
        # Struct support
        self.IS_STRUCT = False
        self.IS_STRUCT_PACKED = False

        self.Parse()

    def ParseStereotypesAbstractAndDocs(self):
        """ Parse stereotypes.
            For 'interface' : pure virtual...
            For 'autogen' : the class will already be generated (i.e. statemachine, or protocol)...so don't regenerate.

            If the user has marked the class as 'abstract' this does the same as having the 'interface' stereotype
        """
        for k, v in self.dict_from_BLOB_STRING.items():
            if k.lower().find('child') > -1:
                for kk, vv in v.items():
                    if kk.lower().find('stereotype') > -1:
                        stereotype = self.parent_classDiagram.table_vppmodelelements.GetModelElement(vv)
                        if stereotype.NAME.lower().find("interface") != -1:
                            self.PURE_VIRTUAL_INTERFACE = True
                        elif stereotype.NAME.lower().find("autogen") != -1:
                            self.AUTOGEN = True
                        elif stereotype.NAME.lower().find("enumeration") != -1:
                            self.IS_ENUM = True
                        elif stereotype.NAME.lower().find("struct") != -1:
                            self.IS_STRUCT = True
                            if stereotype.NAME.lower().find("packed") != -1:
                                self.IS_STRUCT_PACKED = True
                        else:
                            print("Class : unhandled stereotype : " + stereotype.NAME)
                    elif kk.lower().find('abstract') > -1:
                        self.PURE_VIRTUAL_INTERFACE = True
                    elif kk.lower().find('documentation_plain') > -1:
                        self.USER_COMMENTS = vv
                    elif kk.lower().find('child') > -1:
                        if self.IS_ENUM:
                            # Get enumeration literals...
                            if 'type' in vv:
                                if vv['type'].lower().strip() == 'enumerationliteral':
                                    self.ENUM_LITERALS.append(vv['name'].strip())

    def ParseAttributes(self):
        for k, v in self.dict_from_BLOB_STRING.items():
            if k.lower().find('child') > -1:
                for kk, vv in v.items():
                    if kk.lower().find('child') > -1:
                        if vv:  # An empty dictionary evaluates to False
                            if "attribute" == vv['type'].lower():
                                op = ClassAttribute(vv, self.parent_classDiagram.table_vppmodelelements)
                                self.ATTRIBUTES.append(op)

    def ParseOperations(self):
        for k, v in self.dict_from_BLOB_STRING.items():
            if k.lower().find('child') > -1:
                for kk, vv in v.items():
                    if kk.lower().find('child') > -1:
                        if vv:  # An empty dictionary evaluates to False
                            if "operation" == vv['type'].lower():
                                op = ClassOperation(vv, self.parent_classDiagram.table_vppmodelelements)
                                self.OPERATIONS.append(op)

    def Parse(self):
        self.dict_from_BLOB_STRING = vppfs.ParseBLOB_Recursive(self.BLOB_STRING)
        # To look at the structure
        # vppfs.Print_RecursiveParsed_BLOB(self.dict_from_BLOB_STRING)
        self.ParseStereotypesAbstractAndDocs()
        self.ParseOperations()
        self.ParseAttributes()
        #print(self.BLOB_STRING)

    # Helpers...
    def GetNotForwardDeclarableNonPrimitiveTypesLinkedToThis(self):
        """
        Get all classes (fully qualified i.e. "NS1::NS2::CClass") for non-primitive types,
        related to this class, that can not be forward declared.

        Need to inspect
            - inheritance
            - the attributes
            - the operation parameters
            - operation return type
            - associations

        @return: set
        """
        filterValue = set()
        # INHERITANCE
        for id, inheritance in self.parent_classDiagram.inheritence.items():
            if inheritance.CLASS_TO_ID.find(self.ID) > -1:
                if not IsTypePrimitive(inheritance.CLASS_FROM):  # -> YOU never know who may do such a thing.
                    filterValue.add(inheritance.CLASS_FROM)
        # ATTRIBUTES
        for attr in self.ATTRIBUTES:
            if not IsTypePrimitive(attr.TYPE):
                if not IsTypePointerOrRef(attr.TYPE_MODIFIER):
                    filterValue.add(attr.TYPE)
        # OPERATION PARAMS and RETURN TYPE
        for oper in self.OPERATIONS:
            # Operation params
            for params in oper.PARAMETERS:
                modifier = params['modifier']
                type = params['type']
                if not IsTypePrimitive(type):
                    if not IsTypePointerOrRef(modifier):
                        filterValue.add(type)
            # return type
            if not IsTypePrimitive(oper.RETURN_TYPE):
                if not IsTypePointerOrRef(oper.RETURN_TYPE_MODIFIER):
                    filterValue.add(oper.RETURN_TYPE)
        # Associations
        for id, assoc in self.parent_classDiagram.associations.items():
            # This is the only association where it will not be a pointer.
            if assoc.CLASS_FROM_ID == self.ID:
                if assoc.TYPE.lower().find("composition") > -1:
                    if not IsTypePrimitive(assoc.CLASS_TO):  # -> YOU never know who may do such a thing.
                        filterValue.add(assoc.CLASS_TO)
        return filterValue

    def GetForwardDeclarableNonPrimitiveTypesLinkedToThis(self):
        """
        Get all classes (full qualified i.e. "NS1::NS2::CClass") for non-primitive types,
        related to this class, that can be forward declared.

         Need to inspect
            - the attributes
            - the operation parameters
            - operation return type
            - associations

        NOTE: take into account the fact that a ref/ptr AND value can declared...

        @return: set
        """
        filterPtrOrRef = set()
        filterValue = set()
        # ATTRIBUTES
        for attr in self.ATTRIBUTES:
            if not IsTypePrimitive(attr.TYPE):
                if IsTypePointerOrRef(attr.TYPE_MODIFIER):
                    filterPtrOrRef.add(attr.TYPE)
                else:
                    filterValue.add(attr.TYPE)
        # OPERATION PARAMS and RETURN TYPE
        for oper in self.OPERATIONS:
            # Operation params
            for params in oper.PARAMETERS:
                modifier = params['modifier']
                type = params['type']
                if not IsTypePrimitive(type):
                    if IsTypePointerOrRef(modifier):
                        filterPtrOrRef.add(type)
                    else:
                        filterValue.add(type)
            # return type
            if not IsTypePrimitive(oper.RETURN_TYPE):
                if IsTypePointerOrRef(oper.RETURN_TYPE_MODIFIER):
                    filterPtrOrRef.add(oper.RETURN_TYPE)
                else:
                    filterValue.add(oper.RETURN_TYPE)
        # Associations
        for id, assoc in self.parent_classDiagram.associations.items():
            # These are the only association where it will be a pointer.
            if assoc.TYPE.lower().find("association") > -1:  # This is the only one where it is bidirectional
                if assoc.CLASS_FROM_ID == self.ID:
                    if not IsTypePrimitive(assoc.CLASS_TO):  # -> YOU never know who may do such a thing.
                        filterPtrOrRef.add(assoc.CLASS_TO)
                elif assoc.CLASS_TO_ID == self.ID:
                    if not IsTypePrimitive(assoc.CLASS_FROM):  # -> YOU never know who may do such a thing.
                        filterPtrOrRef.add(assoc.CLASS_FROM)
            if assoc.TYPE.lower().find("aggregation") > -1:  # This is the only one where it is bidirectional
                if assoc.CLASS_FROM_ID == self.ID:
                    if not IsTypePrimitive(assoc.CLASS_TO):  # -> YOU never know who may do such a thing.
                        filterPtrOrRef.add(assoc.CLASS_TO)

        # In this case, if the type is in both, the value makes it not-forward-declarable...so remove.
        for i in filterValue:
            if i in filterPtrOrRef:
                filterPtrOrRef.remove(i)

        return filterPtrOrRef

    def GetAssociationsAsListOfAttributesPerVisibility(self, visibility="all"):
        """
        Returns a [] of Associations for this class as ClassAttributes for the desired visibility
         - public
         - protected
         - private
         - all (public and protected and private)
        @param visibility: string
        """
        result = []
        for id, assoc in self.parent_classDiagram.associations.items():
            # This is the only association where it will not be a pointer.
            if assoc.TYPE.lower().find("composition") > -1:
                if assoc.CLASS_FROM_ID == self.ID:
                    if visibility.lower().strip() == assoc.CLASS_FROM_VISIBILITY.lower().strip() or visibility.lower().strip() == "all":
                        if not IsTypePrimitive(assoc.CLASS_TO):  # -> YOU never know who may do such a thing.
                            c = ClassAttribute(None, None)
                            name = CleanName(assoc.NAME)
                            if not name:
                                name = "m_" + assoc.CLASS_TO.replace(self.NAMESPACE + "::", "").replace("::", "_")
                            c.From(name, assoc.CLASS_TO, assoc.CLASS_TO_ID, "", assoc.CLASS_FROM_MULTIPLICITY, assoc.CLASS_FROM_IS_STATIC, assoc.CLASS_FROM_IS_CONST, assoc.CLASS_FROM_VISIBILITY, assoc.CLASS_FROM_HAS_GETTER, assoc.CLASS_FROM_HAS_SETTER, assoc.USER_COMMENTS)
                            c.VISIBILITY = assoc.CLASS_FROM_VISIBILITY
                            result.append(c)
            # These are the only association where it will be a pointer.
            if assoc.TYPE.lower().find("association") > -1:  # This is the only one where it is bidirectional
                # This is also the only type of association whereby a child can contain a parent, and be itself (i.e. composite).
                is_composite = assoc.CLASS_FROM_ID == self.ID and assoc.CLASS_TO_ID == self.ID
                if assoc.CLASS_FROM_ID == self.ID:
                    if visibility.lower().strip() == assoc.CLASS_FROM_VISIBILITY.lower().strip() or visibility.lower().strip() == "all":
                        if not IsTypePrimitive(assoc.CLASS_TO):  # -> YOU never know who may do such a thing.
                            c = ClassAttribute(None, None)
                            name = CleanName(assoc.NAME)
                            if not name or is_composite:
                                name = "m_" + ("child_" if is_composite else "") + assoc.CLASS_TO.replace(self.NAMESPACE + "::", "").replace("::", "_")
                            c.From(name, assoc.CLASS_TO, assoc.CLASS_TO_ID, "*", assoc.CLASS_FROM_MULTIPLICITY, assoc.CLASS_FROM_IS_STATIC, assoc.CLASS_FROM_IS_CONST, assoc.CLASS_FROM_VISIBILITY, assoc.CLASS_FROM_HAS_GETTER, assoc.CLASS_FROM_HAS_SETTER, assoc.USER_COMMENTS)
                            c.VISIBILITY = assoc.CLASS_FROM_VISIBILITY
                            result.append(c)
                if assoc.CLASS_TO_ID == self.ID:
                    if visibility.lower().strip() == assoc.CLASS_TO_VISIBILITY.lower().strip() or visibility.lower().strip() == "all":
                        if not IsTypePrimitive(assoc.CLASS_FROM):  # -> YOU never know who may do such a thing.
                            c = ClassAttribute(None, None)
                            name = CleanName(assoc.NAME)
                            if not name or is_composite:
                                name = "m_" + ("parent_" if is_composite else "") + assoc.CLASS_FROM.replace(self.NAMESPACE + "::", "").replace("::", "_")
                            c.From(name, assoc.CLASS_FROM, assoc.CLASS_FROM_ID, "*", assoc.CLASS_TO_MULTIPLICITY, assoc.CLASS_TO_IS_STATIC, assoc.CLASS_TO_IS_CONST, assoc.CLASS_TO_VISIBILITY, assoc.CLASS_TO_HAS_GETTER, assoc.CLASS_TO_HAS_SETTER, assoc.USER_COMMENTS)
                            c.VISIBILITY = assoc.CLASS_TO_VISIBILITY
                            result.append(c)
            if assoc.TYPE.lower().find("aggregation") > -1:  # This is the only one where it is bidirectional
                if assoc.CLASS_FROM_ID == self.ID:
                    if visibility.lower().strip() == assoc.CLASS_FROM_VISIBILITY.lower().strip() or visibility.lower().strip() == "all":
                        if not IsTypePrimitive(assoc.CLASS_TO):  # -> YOU never know who may do such a thing.
                            c = ClassAttribute(None, None)
                            name = CleanName(assoc.NAME)
                            if not name:
                                name = "m_" + assoc.CLASS_TO.replace(self.NAMESPACE + "::", "").replace("::", "_")
                            c.From(name, assoc.CLASS_TO, assoc.CLASS_TO_ID, "*", assoc.CLASS_FROM_MULTIPLICITY, assoc.CLASS_FROM_IS_STATIC, assoc.CLASS_FROM_IS_CONST, assoc.CLASS_FROM_VISIBILITY, assoc.CLASS_FROM_HAS_GETTER, assoc.CLASS_FROM_HAS_SETTER, assoc.USER_COMMENTS)
                            c.VISIBILITY = assoc.CLASS_FROM_VISIBILITY
                            result.append(c)
        return result

    def GetContainerMultiplicityType(self, MULTIPLICITY):
        """
        Will determine the container type based on multiplicity.

        @param MULTIPLICITY:
        @return: 'vector', 'array' or 'none'
        """
        if MULTIPLICITY.find("*") != -1:  # vector
            return 'vector'
        elif MULTIPLICITY.find("0..1") != -1:
            return 'none'
        elif MULTIPLICITY.find("0..*") != -1:  # vector
            return 'vector'
        elif MULTIPLICITY.find("1..*") != -1:  # vector
            return 'vector'
        elif MULTIPLICITY.find("..") > -1:
            numbers = MULTIPLICITY.split("..")
            try:  # Array
                i = int(numbers[-1])
                return 'array:' + str(i)
            except ValueError:  # vector
                return 'vector'
        elif MULTIPLICITY.find("0") != -1:
            return 'none'
        elif MULTIPLICITY.find("1") != -1:
            return 'none'
        else:  # lastly...perhaps someone put an actual value here...then its an array.
            try:  # Array
                i = int(MULTIPLICITY)
                return 'array:' + str(i)
            except ValueError:
                return 'none'
        return 'none'

    def DoAttributesAssociationsReturnTypesOrFunctionParametersRequireVector(self):
        """
        @return: True or False if any attributes, associations or operation-parameters use a vector (based on multiplicity)
        """
        has_Vector = False
        for attr in self.ATTRIBUTES:
            if not has_Vector and self.GetContainerMultiplicityType(attr.MULTIPLICITY).find("vector") > -1:
                has_Vector = True

        if not has_Vector:
            associations_as_attributes = self.GetAssociationsAsListOfAttributesPerVisibility("all")
            for attr in associations_as_attributes:
                if not has_Vector and self.GetContainerMultiplicityType(attr.MULTIPLICITY).find("vector") > -1:
                    has_Vector = True

        if not has_Vector:
            for oper in self.OPERATIONS:
                for param in oper.PARAMETERS:
                    if not has_Vector and self.GetContainerMultiplicityType(param['multiplicity']).find("vector") > -1:
                        has_Vector = True;
                # Search for <!#!>
                if oper.RETURN_TYPE_MODIFIER.strip().find("[]") > -1:
                    has_Vector = True

        # Also look at any members from interfaces !
        for id, inheritance in self.parent_classDiagram.inheritence.items():
            if inheritance.CLASS_TO_ID.find(self.ID) > -1:
                if inheritance.IS_REALIZATION:# or REALIZING_CLASS:  # or REALIZING_CLASS -> if a pure virtual interface inherits another pure virtual interface, and the top child class that inherits them all wants to realize, then all should be realized
                    # get the class we are realizing...
                    realizeObj = self.parent_classDiagram.classes[inheritance.CLASS_FROM_ID]
                    if realizeObj.PURE_VIRTUAL_INTERFACE:
                        has_Vector = has_Vector or realizeObj.DoAttributesAssociationsReturnTypesOrFunctionParametersRequireVector()

        return has_Vector


'''----------------------------------------------------------------------'''


class ClassDiagram:
    """ This class represents a class diagram for a particular diagram of 'name' and 'id'.
        The name and the id of the diagram are passed into the constructor.
        The diagram elements from the project are also passed into the constructor (from the diagram elements table)
        as well as the model elements table.

        This class will use the diagram elements to lookup all the necessary model elements from the model elements table
        and create a ??? table from that.
    """

    def __init__(self, diagramName, diagramID, diagramElements, table_vppmodelelements):
        self.name = diagramName
        self.id = diagramID
        self.elements = diagramElements
        self.table_vppmodelelements = table_vppmodelelements

        self.packages = {}
        self.classes = {}
        self.associations = {}
        self.inheritence = {}
        self.usages = {}
        self.LoadAndTest()

    ''' Will load the table and hope to catch any changes (for future proofing) if VP decide to change their database. '''

    def LoadAndTest(self):
        for i in self.elements:
            vppmodelelement = self.table_vppmodelelements.GetModelElement(i.MODEL_ELEMENT_ID)
            #print("|| ", vppmodelelement.MODEL_TYPE)
            if vppmodelelement.MODEL_TYPE == 'Class':
                self.classes[vppmodelelement.ID] = Class(vppmodelelement, self)
            elif vppmodelelement.MODEL_TYPE == 'Package':
                self.packages[vppmodelelement.ID] = Package(vppmodelelement)
            elif vppmodelelement.MODEL_TYPE == 'Association':
                self.associations[vppmodelelement.ID] = Association(vppmodelelement, self.table_vppmodelelements)
            elif vppmodelelement.MODEL_TYPE == 'Realization' or vppmodelelement.MODEL_TYPE == 'Generalization':
                self.inheritence[vppmodelelement.ID] = Inheritance(vppmodelelement, self.table_vppmodelelements, vppmodelelement.MODEL_TYPE == 'Realization')
            elif vppmodelelement.MODEL_TYPE == 'Usage':
                # self.usages[vppmodelelement.ID] = Usage(vppmodelelement)
                pass
            else:
                print("Class Diagram : unhandled model-element type : ", vppmodelelement.MODEL_TYPE)

        # UML Packages are namespaces. The can be nested. This supports nesting.
        # Iterate through all packages, see what classes are in them...these will be the fully qualified (nested) range of IDS,
        # and build the namespace for the class.
        # # # > A side note : the whole idea of de-serializing packages seems moot...as the PARENT_ID member of the class objects, when embedded in a package...will be the ID for the package...
        # # # > ...and I guess the same for a package...and so that could be all that is used, by looking up and looking up etc.
        # # # > HOWEVER, if someone used another shape as a package...then that should be the namespace name...and we might have to do this.
        for _id, package in self.packages.items():
            for pclassid in package.CLASSES_IN_PACKAGE:
                fullyQualifiedNestedNamespacesAndClass = pclassid.split(':')
                namespace = ""
                for i in range(len(fullyQualifiedNestedNamespacesAndClass) - 1):
                    namespace = namespace + self.packages[fullyQualifiedNestedNamespacesAndClass[i]].NAME + "::"
                namespace = namespace.rstrip(":")  # Remove the last '::'

                for classid, _class in self.classes.items():
                    if classid == fullyQualifiedNestedNamespacesAndClass[len(fullyQualifiedNestedNamespacesAndClass) - 1]:
                        _class.NAMESPACE = namespace
                        break

        # For some reason, nest typeid's for inheritance are not fully qualified...fix this AFTER fixing namespace names.
        for id, inh in self.inheritence.items():
            inh.PostProjectParseFix(self)

    ''' Returns a dictionary of 'namespace : set()' dependencies where the set is a set of namespaces the 'key' namespace depends on.
    '''

    def GetNamespaceDependencies(self):
        result = {}
        for class_uid, classobj in self.classes.items():
            namespace = classobj.NAMESPACE
            if not namespace in result:
                result[namespace] = set()
            # These are sets. Dont add a dependancy to oneself...
            set1 = classobj.GetNotForwardDeclarableNonPrimitiveTypesLinkedToThis()
            set2 = classobj.GetForwardDeclarableNonPrimitiveTypesLinkedToThis()
            for s in set1:
                if s.find(namespace) == -1:
                    result[namespace].add(s.rpartition("::")[0])
            for s in set2:
                if s.find(namespace) == -1:
                    result[namespace].add(s.rpartition("::")[0])
        return result

''' USE THIS FUNCTION to extract a named class diagram from a VPP file.
    input : class diagram name
    input : path to the file
'''
def ExtractClassDiagram(classdiagramname, path_to_vpp):
    vppdiagrams = vppfs.VPPDiagrams(path_to_vpp)
    vppdiagramelements = vppfs.VPPDiagramElements(path_to_vpp)
    vppmodelelements = vppfs.VPPModelElements(path_to_vpp)
    elements = vppdiagramelements.GetDiagramElements(vppdiagrams.GetIDFromClassDiagramName(classdiagramname))
    class_diagram = ClassDiagram(classdiagramname, vppdiagrams.GetIDFromClassDiagramName(classdiagramname), elements, vppmodelelements)
    return class_diagram

''' USE THIS FUNCTION to extract ALL class diagrams from a VPP file.
    input : path to the file
'''
def ExtractAllClassDiagrams(path_to_vpp):
    vppdiagrams = vppfs.VPPDiagrams(path_to_vpp)
    vppdiagramelements = vppfs.VPPDiagramElements(path_to_vpp)
    vppmodelelements = vppfs.VPPModelElements(path_to_vpp)
    class_diagrams = []
    for id, name in vppdiagrams.class_diagrams.items():
        elements = vppdiagramelements.GetDiagramElements(id)
        class_diagram = ClassDiagram(name, id, elements, vppmodelelements)
        class_diagrams.append(class_diagram)
    return class_diagrams


import unittest
import shutil
import os
from kojen.LanguagePython import LanguagePython
from kojen.LanguageCPP import LanguageCPP
from kojen.LanguageCsharp import LanguageCsharp
from kojen.smgen import CStateMachineGenerator
from kojen.kojentypes import Interface, Struct, Message, MessageHeader

class TestFeatures(unittest.TestCase):

    workingfolder = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test")

    @classmethod
    def setUpClass(cls):
        os.makedirs(cls.workingfolder, exist_ok=True)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.workingfolder)

    @classmethod
    def create_template_file(cls, list_of_lines):
        assert os.path.isdir(cls.workingfolder)
        with open(os.path.join(cls.workingfolder, "file.t"), 'w') as f:
            for l in list_of_lines:
                f.write("%s\n" % l)

    @classmethod
    def read_lines_of_output_file(cls):
        path = os.path.join(cls.workingfolder, "file.t")
        assert os.path.isfile(path)
        with open(path, 'r') as f:
            return f.readlines()

    @classmethod
    def do_magic(cls, input, interface, tt = [], language=LanguagePython()):
        TestFeatures.create_template_file(input)
        smgenerator = CStateMachineGenerator(cls.workingfolder, cls.workingfolder, interface, language, "", "", "")
        smgenerator.Generate(tt, "", "", "", False)
        return TestFeatures.read_lines_of_output_file()

    def test_event_custom_params_nosignature(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<EVENTSIGNATURE::p1, p2, p3,>>>")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("somestruct")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0]  , "p1, p2, p3\n")

    def test_event_custom_params_signature1(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<EVENTSIGNATURE::p1, p2, p3,>>>")
        input.append("<<<EVENTSIGNATUREWITHDEFAULTS::p1, p2, p3,>>>")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga","bungaBunga", 0x66)
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "binga, p1, p2, p3\n") # Python
        self.assertEqual(output[1], "binga, p1, p2, p3\n")

    def test_event_custom_params_signature2(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<SIGNATURE>>>")
        input.append("<<<SIGNATUREWITHDEFAULTS>>>")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga","bungaBunga", "0x66")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "binga\n") # Python
        self.assertEqual(output[1], "binga=0x66\n")  # Python

    def test_event_custom_params_signature3(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("(some, <<<SIGNATURE>>>)")
        input.append("(some, <<<SIGNATUREWITHDEFAULTS>>>)")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("s")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "(some)\n") # Python
        self.assertEqual(output[1], "(some)\n")

    def test_event_custom_params_signature4(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("(<<<SIGNATURE>>> , other)")
        input.append("(<<<SIGNATUREWITHDEFAULTS>>> , other)")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("s")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "(other)\n") # Python
        self.assertEqual(output[1], "(other)\n")

    def test_event_custom_params_signature5(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("(some, <<<SIGNATURE>>>)")
        input.append("(some, <<<SIGNATUREWITHDEFAULTS>>>)")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("s")
        s.AddType("binga","bungaBunga", "0x66")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "(some, binga)\n") # Python
        self.assertEqual(output[1], "(some, binga=0x66)\n")

    def test_event_custom_params_signature6(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("(<<<SIGNATURE>>>, yo)")
        input.append("(<<<SIGNATUREWITHDEFAULTS>>>, yo)")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("s")
        s.AddType("binga","bungaBunga", "0x66")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "(binga, yo)\n") # Python
        self.assertEqual(output[1], "(binga=0x66, yo)\n")

    def createMsgInputWithDefaults(self):
        input = []
        input.append("<<<PER_MSG_BEGIN>>>")
        input.append("<<<SIGNATURE>>>")
        input.append("<<<SIGNATUREWITHDEFAULTS>>>")
        input.append("<<<PER_MSG_END>>>")
        return input
    def createIfNestedMsgWithDefalts(self, allDefaults = True):
        s = Struct("s")
        s.AddType("myApple", "apple", 1)
        s.AddType("myOrange", "orange", 2)
        s.AddType("myBanana", "banana", 3)
        s1 = Struct("FruitSalad")
        s1.AddStruct("myFruits", s)
        s1.AddType("myCustard", "custard", 4 if allDefaults else None)
        s2 = Struct("Breakfast")
        s2.AddStruct("myEdibles", s1)
        s2.AddType("myCoffee", "coffee", 5)
        s2.AddType("myOJ", "OJ", 6 if allDefaults else None)
        m = Message('ThisMorning', 0x02)
        m.AddStruct('myBreakfast', s2)
        m.AddType('myKnife', 'knife', 7)
        m.AddType('myFork', 'fork', 8 if allDefaults else None)
        m.AddType('myNapkin', 'napkin', 9 if allDefaults else None)
        i = Interface('')
        i.AddStruct(s)
        i.AddStruct(s1)
        i.AddStruct(s2)
        i.AddMessage(m)
        return i

    def test_event_params_signature_all_defaults(self):
        input = self.createMsgInputWithDefaults()
        i = self.createIfNestedMsgWithDefalts(True)
        r = {}
        r[LanguageCPP()] = [
            'Breakfast const& myBreakfast, knife myKnife, fork myFork, napkin myNapkin\n',
            'Breakfast const& myBreakfast={{{1,2,3},4},5,6}, knife myKnife=7, fork myFork=8, napkin myNapkin=9\n'
        ]
        #r[LanguagePython()] = None
        r[LanguageCsharp()] = [
            'Breakfast myBreakfast, knife myKnife, fork myFork, napkin myNapkin\n',
            'Breakfast myBreakfast={{{1,2,3},4},5,6}, knife myKnife=7, fork myFork=8, napkin myNapkin=9\n'
        ]

        for lan, res in r.items():
            output = TestFeatures.do_magic(input, i, [], lan)
            self.assertEqual(len(output), 2)
            self.assertEqual(output[0], res[0])
            self.assertEqual(output[1], res[1])

    def test_event_params_signature_some_defaults(self):
        input = self.createMsgInputWithDefaults()
        i = self.createIfNestedMsgWithDefalts(False)
        r = {}
        r[LanguageCPP()] = [
            'Breakfast const& myBreakfast, knife myKnife, fork myFork, napkin myNapkin\n',
            'Breakfast const& myBreakfast={{{1,2,3},{}},5,{}}, knife myKnife=7, fork myFork={}, napkin myNapkin={}\n'
        ]
        #r[LanguagePython()] = None
        r[LanguageCsharp()] = [
            'Breakfast myBreakfast, knife myKnife, fork myFork, napkin myNapkin\n',
            'Breakfast myBreakfast={{{1,2,3},{}},5,{}}, knife myKnife=7, fork myFork={}, napkin myNapkin={}\n'
        ]

        for lan, res in r.items():
            output = TestFeatures.do_magic(input, i, [], lan)
            self.assertEqual(len(output), 2)
            self.assertEqual(output[0], res[0])
            self.assertEqual(output[1], res[1])

    def test_event_members_declare(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<MEMBERSDECLARE>>>")
        input.append("<<<PER_EVENT_END>>>")

        s = Struct("s")
        s.AddType("binga", "bungaBunga", "0x66")
        s.AddType("bonga", "bangaBanga")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)#, [], LanguageCsharp())
        self.assertEqual(len(output), 2)
        self.assertEqual(output[0].strip(' '), "binga = 0x66 # bungaBunga\n")
        self.assertEqual(output[1].strip(' '), "bonga = None # bangaBanga\n")

    def test_event_members_instantiate_custom_name(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<EVENTMEMBERSLITEINSTANTIATE::hello>>>")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga", "bungaBunga")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0], "hello.binga = binga\n")

    def test_event_members_instantiate_no_custom_name(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<MEMBERSLITEINSTANTIATE>>>")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga", "bungaBunga")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i)

        self.assertEqual(len(output), 1)
        self.assertEqual(output[0], "data.binga = binga\n")

    def test_transitionsperguard_tags(self):
        input = []
        input.append("<<<PER_STATETRANSITION_BEGIN>>>")
        input.append("<<<PER_EVENTTRANSITION_BEGIN>>>")
        input.append("<<<PER_GUARDTRANSITION_BEGIN>>>")
        input.append("<<<GUARDNAME::No Guard>>>")
        #input.append("<<<EVENTNAME::No Event>>>") Hmmm ... need to think more on this.
        input.append("<<<EVENTNAME>>>")
        input.append("<<<STATENAMEIFNEXTSTATE::No Next State>>>")
        input.append("<<<ACTIONNAME::No Action>>>")
        input.append("<<<NEXTSTATENAME::No Next State>>>")
        input.append("<<<PER_GUARDTRANSITION_END>>>")
        input.append("<<<PER_EVENTTRANSITION_END>>>")
        input.append("<<<PER_STATETRANSITION_END>>>")
        tt = [['S1', 'Do', 'S1', 'A1', 'G1']]
        s = Struct("somestruct")
        s.AddType("binga", "bungaBunga")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i, tt)

        self.assertEqual(len(output), 5)
        self.assertEqual(output[0], "G1\n")
        self.assertEqual(output[1], "Do\n")
        self.assertEqual(output[2], "S1\n")
        self.assertEqual(output[3], "A1\n")
        self.assertEqual(output[4], "S1\n")

    def test_transitionsperguard_tags_with_custom_defaults(self):
        input = []
        input.append("<<<PER_STATETRANSITION_BEGIN>>>")
        input.append("<<<PER_EVENTTRANSITION_BEGIN>>>")
        input.append("<<<PER_GUARDTRANSITION_BEGIN>>>")
        input.append("<<<GUARDNAME::No Guard>>>")
        input.append("<<<EVENTNAME::No Event>>>")
        input.append("<<<STATENAMEIFNEXTSTATE::No Next State>>>")
        input.append("<<<ACTIONNAME::No Action>>>")
        input.append("<<<NEXTSTATENAME::No Next State>>>")
        input.append("<<<PER_GUARDTRANSITION_END>>>")
        input.append("<<<PER_EVENTTRANSITION_END>>>")
        input.append("<<<PER_STATETRANSITION_END>>>")
        tt = [['S1', 'Do', 'None', 'None', 'None']]
        s = Struct("somestruct")
        s.AddType("binga", "bungaBunga")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i, tt)

        self.assertEqual(len(output), 5)
        self.assertEqual(output[0], "No Guard\n")
        self.assertEqual(output[1], "No Event\n")
        self.assertEqual(output[2], "No Next State\n")
        self.assertEqual(output[3], "No Action\n")
        self.assertEqual(output[4], "No Next State\n")

    def test_events(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<EVENTNAME>>>")
        input.append("<<<PER_EVENT_END>>>")
        tt = [['S1', "how ya doin'", 'None', 'None', 'None']]
        s = Struct("hello")
        i = Interface('')
        i.AddStruct(s)

        output = TestFeatures.do_magic(input, i, tt)

        self.assertEqual(len(output), 2)
        self.assertEqual(output[1], "hello\n")
        self.assertEqual(output[0], "how ya doin'\n")

    def test_aggregate_initialization_events(self):
        input = []
        input.append("<<<PER_EVENT_BEGIN>>>")
        input.append("<<<AGGREGATEINITIALIZATION>>>")
        input.append("<<<PER_EVENT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga", "bool")
        s.AddType("bunga","size_t")
        s2 = Struct("somestruct2")
        s2.AddType("bla", "bool")
        s2.AddType("blabla", "size_t")
        i = Interface('')
        i.AddStruct(s)
        i.AddStruct(s2)

        output = TestFeatures.do_magic(input, i, [], LanguageCPP())

        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "{binga, bunga}\n")
        self.assertEqual(output[1], "{bla, blabla}\n")

    def test_pyattr(self):
        # param, no default
        input = []
        input.append("<<<PER_STRUCT_BEGIN>>>")
        input.append("<<<PyAttr::ninja>>>")
        input.append("<<<PER_STRUCT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga", "bool")
        s.AddType("bunga", "size_t")
        s.ninja = "here"
        s2 = Struct("somestruct2")
        s2.AddType("bla", "bool")
        s2.AddType("blabla", "size_t")
        s2.ninja = 36
        i = Interface('')
        i.AddStruct(s)
        i.AddStruct(s2)
        output = TestFeatures.do_magic(input, i, [], LanguageCPP())
        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "here\n")
        self.assertEqual(output[1], "36\n")
        # multiple param, no default
        input = []
        input.append("<<<PER_STRUCT_BEGIN>>>")
        input.append("<<<PyAttr::ninja>>> is a big fat <<<PyAttr::ninja>>>")
        input.append("<<<PER_STRUCT_END>>>")
        output = TestFeatures.do_magic(input, i, [], LanguageCPP())
        self.assertEqual(len(output), 2)
        self.assertEqual(output[0], "here is a big fat here\n")
        self.assertEqual(output[1], "36 is a big fat 36\n")
        # no param
        input = []
        input.append("<<<PER_STRUCT_BEGIN>>>")
        input.append("<<<PyAttr::ninja>>>")
        input.append("<<<PER_STRUCT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga", "bool")
        s.AddType("bunga", "size_t")
        s2 = Struct("somestruct2")
        s2.AddType("bla", "bool")
        s2.AddType("blabla", "size_t")
        i = Interface('')
        i.AddStruct(s)
        i.AddStruct(s2)
        output2 = TestFeatures.do_magic(input, i, [], LanguageCPP())
        self.assertEqual(len(output2), 0)
        # param, default
        input = []
        input.append("<<<PER_STRUCT_BEGIN>>>")
        input.append("<<<PyAttr::ninja::yoyo>>>")
        input.append("<<<PER_STRUCT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga", "bool")
        s.AddType("bunga", "size_t")
        s2 = Struct("somestruct2")
        s2.AddType("bla", "bool")
        s2.AddType("blabla", "size_t")
        i = Interface('')
        i.AddStruct(s)
        i.AddStruct(s2)
        output3 = TestFeatures.do_magic(input, i, [], LanguageCPP())
        self.assertEqual(len(output3), 2)
        self.assertEqual(output3[0], "yoyo\n")
        self.assertEqual(output3[1], "yoyo\n")

    def test_Docs(self):
        input = []
        input.append("<<<PER_STRUCT_BEGIN>>>")
        input.append("  *  *  <<<DOCUMENTATION>>>")
        input.append("<<<PER_STRUCT_END>>>")
        s = Struct("somestruct")
        s.AddType("binga", "bool")
        s.AddType("bunga", "size_t")
        s.SetDocumentation("line1\nline2\nline3\n")
        s2 = Struct("somestruct2")
        s2.AddType("bla", "bool")
        s2.AddType("blabla", "size_t")
        s2.SetDocumentation("line4\nline5\nline6")
        i = Interface('')
        i.AddStruct(s)
        i.AddStruct(s2)
        # param, no default
        output = TestFeatures.do_magic(input, i, [], LanguageCPP())
        self.assertEqual(len(output), 6)
        self.assertEqual(output[0], "  *  *  line1\n")
        self.assertEqual(output[1], "  *  *  line2\n")
        self.assertEqual(output[2], "  *  *  line3\n")
        self.assertEqual(output[3], "  *  *  line4\n")
        self.assertEqual(output[4], "  *  *  line5\n")
        self.assertEqual(output[5], "  *  *  line6\n")
    '''
    def test_split(self):
        s = 'hello world'
        self.assertEqual(s.split(), ['hello', 'world'])
        # check that s.split fails when the separator is not a string
        with self.assertRaises(TypeError):
            s.split(2)
    '''

if __name__ == '__main__':
    unittest.main()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = 'eugene'

'''

    MIT License

    Copyright (c) 2015 Eugene Grobbelaar (email : koh.jaen@yahoo.de)

    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included in all
    copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
    SOFTWARE.

'''
__TAG_AUTHOR__                      = '<<<AUTHOR>>>'
__TAG_GROUP__                       = '<<<GROUP>>>'
__TAG_BRIEF__                       = '<<<BRIEF>>>'
__TAG_NAMESPACE__                   = '<<<NAMESPACE>>>'
__TAG_SM_NAME__                     = '<<<STATEMACHINENAME>>>'       # As given
__TAG_SM_NAME_SMALL_CAMEL__         = '<<<stateMachineName>>>'       # camelCaps
__TAG_SM_NAME_UPPER__               = '<<<STATEMACHINENAMEUPPER>>>'  # ALL UPPER
__TAG_CLASS_NAME__                  = '<<<CLASSNAME>>>'
__TAG_PyIFGen_NAME__                = '<<<PYIFGENNAME>>>'
__TAG_ENUMERATIONS__                = '<<<ENUMS>>>'

__TAG_PS_BEGIN__                    = "<<<PER_STATE_BEGIN>>>"
__TAG_PS_END__                      = "<<<PER_STATE_END>>>"

__TAG_PST_BEGIN__                   = "<<<PER_STATETRANSITION_BEGIN>>>"
__TAG_PST_END__                     = "<<<PER_STATETRANSITION_END>>>"
__TAG_PET_BEGIN__                   = "<<<PER_EVENTTRANSITION_BEGIN>>>"
__TAG_PET_END__                     = "<<<PER_EVENTTRANSITION_END>>>"
__TAG_PGT_BEGIN__                   = "<<<PER_GUARDTRANSITION_BEGIN>>>"
__TAG_PGT_END__                     = "<<<PER_GUARDTRANSITION_END>>>"

__TAG_PE_BEGIN__                    = "<<<PER_EVENT_BEGIN>>>"
__TAG_PE_END__                      = "<<<PER_EVENT_END>>>"

__TAG_PA_BEGIN__                    = "<<<PER_ACTION_BEGIN>>>"
__TAG_PA_END__                      = "<<<PER_ACTION_END>>>"

__TAG_PASIG_BEGIN__                 = "<<<PER_ACTION_SIGNATURE_BEGIN>>>"
__TAG_PASIG_END__                   = "<<<PER_ACTION_SIGNATURE_END>>>"

__TAG_PG_BEGIN__                    = "<<<PER_GUARD_BEGIN>>>"
__TAG_PG_END__                      = "<<<PER_GUARD_END>>>"

__TAG_STATENAME__                   = '<<<STATENAME>>>'            # As given
__TAG_STATENAME_SMALL_CAMEL__       = '<<<stateName>>>'            # camelCaps
__TAG_NEXTSTATENAME__               = '<<<NEXTSTATENAME>>>'        # As given
__TAG_NEXTSTATENAME_SMALL_CAMEL__   = '<<<nextStateName>>>'        # camelCaps
__TAG_STATENAME_IF_NEXTSTATE__      = '<<<STATENAMEIFNEXTSTATE>>>' # As given
__TAG_STATENAME_IF_NEXTSTATE_SMALL_CAMEL__      = '<<<stateNameIfNextState>>>' # As given
__TAG_EVENTNAME__                   = '<<<EVENTNAME>>>'            # As given
__TAG_EVENTNAME_SMALL_CAMEL__       = '<<<eventName>>>'            # camelCaps
__TAG_ACTIONNAME__                  = '<<<ACTIONNAME>>>'           # As given
__TAG_ACTIONNAME_SMALL_CAMEL__      = '<<<actionName>>>'           # camelCaps
__TAG_GUARDNAME__                   = '<<<GUARDNAME>>>'            # As given
__TAG_GUARDNAME_SMALL_CAMEL__       = '<<<guardName>>>'            # camelCaps

__TAG_ABC__                         = '<<<ALPH>>>'
__TAG_123__                         = '<<<NUM>>>'
__TAG_INIT_STATE__                  = '<<<STATE_0>>>'              # As given
__TAG_INIT_STATE_SMALL_CAMEL__      = '<<<state_0>>>'              # camelCaps

__TAG_TTT_BOOST_MSM__               = '<<<TTT_BOOST_MSM>>>'
__TAG_TTT_BOOST_MSMLITE__           = '<<<TTT_BOOST_MSMLITE>>>'
__TAG_TTT_PLANT_UML__               = '<<<TTT_PLANT_UML>>>'
__TAG_TTT_BOOST_SML__               = '<<<TTT_BOOST_SML>>>'
__TAG_TTT_BOOST_SML_ENTRYEXIT__     = '<<<TTT_BOOST_SML_ENTRY_EXIT>>>'

__TAG_DECLSPEC_DLL_EXPORT__         = "<<<DLL_EXPORT>>>"

### CONSOLODATE AND unit test the new recursive features with defaults
__TAG_STRUCTNAME__                  = '<<<STRUCTNAME>>>'        # As given
__TAG_STRUCTNAME_SMALL_CAMEL__      = '<<<structName>>>'        # camelCaps
__TAG_MSGNAME__                     = '<<<MSGNAME>>>'           # As given
__TAG_MSGID__                       = '<<<MSGID>>>'
__TAG_MSGNAME_SMALL_CAMEL__         = '<<<msgName>>>'           # camelCaps
__TAG_PROTOMSGNAME__                = '<<<PROTOMSGNAME>>>'      # As given
__TAG_PROTOMSGNAME_SMALL_CAMEL__    = '<<<protoMsgName>>>'      # camelCaps
__TAG_ATTRIBUTE_TYPE__              = "<<<ATTRIBUTETYPE>>>"
__TAG_ATTRIBUTE_NAME__              = "<<<ATTRIBUTENAME>>>"
__TAG_PAYLOAD_TYPE__                = "<<<PAYLOADTYPE>>>"
__TAG_PAYLOAD_NAME__                = "<<<PAYLOADNAME>>>"
__TAG_PYTHON_ATTR__                 = "<<<PyAttr>>>"
__TAG_DOCUMENTATION__               = "<<<DOCUMENTATION>>>"

__TAG_STRUCT_BEGIN__                = "<<<PER_STRUCT_BEGIN>>>"
__TAG_STRUCT_END__                  = "<<<PER_STRUCT_END>>>"
__TAG_MSG_BEGIN__                   = "<<<PER_MSG_BEGIN>>>"
__TAG_MSG_END__                     = "<<<PER_MSG_END>>>"
__TAG_PROTOMSG_BEGIN__              = "<<<PER_PROTOMSG_BEGIN>>>"
__TAG_PROTOMSG_END__                = "<<<PER_PROTOMSG_END>>>"
###
__TAG_SIGNATURE__                   = "<<<SIGNATURE>>>"
__TAG_SIGNATURE_DEF__               = "<<<SIGNATUREWITHDEFAULTS>>>"
__TAG_MEMBERINST__                  = "<<<MEMBERSINSTANTIATE>>>"
__TAG_LITE_MEMBERINST__             = "<<<MEMBERSLITEINSTANTIATE>>>"
__TAG_MEMBERDECL__                  = "<<<MEMBERSDECLARE>>>"
__TAG_AGGREGATE_INIT__              = "<<<AGGREGATEINITIALIZATION>>>"

# Python2 -> 3 shennanigans...try support both
try:
    from kojentypes import *		# py2
except (ModuleNotFoundError, ImportError) as e:
    from .kojentypes import *		# py3

try:
    from .preservative import *
except (ModuleNotFoundError, ImportError) as e:
    from preservative import *

try:
    from .cgen import *
except (ModuleNotFoundError, ImportError) as e:
    from cgen import *

try:
    from LanguageCPP import LanguageCPP
except  (ModuleNotFoundError, ImportError) as e:
    from .LanguageCPP import LanguageCPP

try:
    from plant import TTToDot
except  (ModuleNotFoundError, ImportError) as e:
    from .plant import TTToDot

import re

# Model that describes a state machine.
class CStateMachineModel:
    def __init__(self):
        self.statemachinename = ""
        self.namespacename    = ""
        self.declspecdllexport = ""
        self.pythoninterfacegeneratorfilename = ""
        self.states           = []
        self.actions          = []
        self.events           = []
        self.guards           = []
        self.actionsignatures = OrderedDict()
        self.transitionsperstate = OrderedDict()

# Transition Table Model uses State Machine Model to generate all code required for a working state machine.
class CTransitionTableModel(CStateMachineModel):
    START_STATE = 0
    EVENT       = 1
    NEXT_STATE  = 2
    ACTION      = 3
    GUARD       = 4

    def __init__(self, tt, nn, smn, dclspc = ""):
        CStateMachineModel.__init__(self)
        self.transition_table   = tt
        self.statemachinename   = smn
        self.namespacename      = nn
        self.declspecdllexport  = dclspc
        tstate  = OrderedDict() # TODO : Since py 3.7 a regular dict is guaranteed to preserve order, and is more performat.
        taction = OrderedDict()
        tevent  = OrderedDict()
        tguard  = OrderedDict()
        self.maxlenSTART_STATE = 0
        self.maxlenEVENT = 0
        self.maxlenACTION = 0
        self.maxlenGUARD = 0

        length = 0
        # Filter
        for tableline in self.transition_table:
            if tableline[self.START_STATE] != "" and tableline[self.START_STATE].lower() != "none":
                tstate[tableline[self.START_STATE]] = 0
                length = len(tableline[self.START_STATE])
                if self.maxlenSTART_STATE < length:
                    self.maxlenSTART_STATE = length
            if tableline[self.NEXT_STATE] != "" and tableline[self.NEXT_STATE].lower() != "none":
                tstate[tableline[self.NEXT_STATE]] = 0
            if tableline[self.EVENT] != "" and tableline[self.EVENT].lower() != "none":
                tevent[tableline[self.EVENT]] = 0
                length = len(tableline[self.EVENT])
                if self.maxlenEVENT < length:
                    self.maxlenEVENT = length
            if tableline[self.ACTION] != "" and tableline[self.ACTION].lower() != "none":
                taction[tableline[self.ACTION]] = 0
                length = len(tableline[self.ACTION])
                if self.maxlenACTION < length:
                    self.maxlenACTION = length
                if not ((tableline[self.ACTION] + tableline[self.EVENT]) in self.actionsignatures):
                    self.actionsignatures[tableline[self.ACTION] + tableline[self.EVENT]] = (tableline[self.ACTION], tableline[self.EVENT])  #, tableline[self.START_STATE],tableline[self.NEXT_STATE]))
            if tableline[self.GUARD] != "" and tableline[self.GUARD].lower() != "none":
                tguard[tableline[self.GUARD]] = 0
                length = len(tableline[self.GUARD])
                if self.maxlenGUARD < length:
                    self.maxlenGUARD = length
        # Populate CStateMachineModel
        for s in tstate:
            self.states.append(s)
        for e in tevent:
            self.events.append(e)
        for a in taction:
            self.actions.append(a)
        for g in tguard:
            self.guards.append(g)
        self.set_transitions_per_state()

    ''' Returns a dictionary of dictionaries of lists of dictionaries.
    
        {'StateName': {'EventName',[{'GuardName':'', 'ActionName':'val'},{}...]}}
        First dictionary key is the 'state'.
        Second dictionary is the 'event' name.
        These contain a list of dictionaries, as multiple of the same events can
        cause different transitions based on guards.
    '''
    def set_transitions_per_state(self):
        for tableline in self.transition_table:
            transition = OrderedDict()
            if tableline[self.ACTION] != "" and tableline[self.ACTION].lower() != "none":
                transition[__TAG_ACTIONNAME__] = tableline[self.ACTION]
                transition[__TAG_ACTIONNAME_SMALL_CAMEL__] = camel_case_small(tableline[self.ACTION])
            if tableline[self.GUARD] != "" and tableline[self.GUARD].lower() != "none":
                transition[__TAG_GUARDNAME__] = tableline[self.GUARD]
                transition[__TAG_GUARDNAME_SMALL_CAMEL__] = camel_case_small(tableline[self.GUARD])
            if tableline[self.NEXT_STATE] != "" and tableline[self.NEXT_STATE].lower() != "none":
                transition[__TAG_STATENAME_IF_NEXTSTATE__] = tableline[self.START_STATE]
                transition[__TAG_STATENAME_IF_NEXTSTATE_SMALL_CAMEL__] = camel_case_small(tableline[self.START_STATE])
                transition[__TAG_NEXTSTATENAME__] = tableline[self.NEXT_STATE]
                transition[__TAG_NEXTSTATENAME_SMALL_CAMEL__] = camel_case_small(tableline[self.NEXT_STATE])

            if tableline[self.START_STATE] != "" and tableline[self.START_STATE].lower() != "none":
                if not tableline[self.START_STATE] in self.transitionsperstate:
                    self.transitionsperstate[tableline[self.START_STATE]] = OrderedDict()

            if tableline[self.EVENT] != "" and tableline[self.EVENT].lower() != "none":
                if not tableline[self.EVENT] in self.transitionsperstate[tableline[self.START_STATE]]:
                    self.transitionsperstate[tableline[self.START_STATE]][tableline[self.EVENT]] = []
                self.transitionsperstate[tableline[self.START_STATE]][tableline[self.EVENT]].append(transition)

    def getfirststate(self):
        if not self.transition_table:
            return "NO TT PRESENT!"
        return self.transition_table[0][0]


class CStateMachineGenerator(CGenerator):

    def __init__(self, inputfiledir, outputfiledir, events_interface=None, language=None, author='Anonymous', group='', brief=''):
        CGenerator.__init__(self,inputfiledir,outputfiledir,language, author, group, brief)
        self.events_interface = events_interface
        self.vpp_filename = ""

    def loadtemplates_firstfiltering(self, smmodel):
        """
        See baseclass implementation. This just prepares the dictionary of things to replace
        for this type of codegeneration.

        @param smmodel:
        @return: cgen.CCodeModel, a dictionary -> {filename,[lines]}
        """

        dict_to_replace_lines = {}
        dict_to_replace_lines[__TAG_SM_NAME_UPPER__] = caps(smmodel.statemachinename)
        dict_to_replace_lines[__TAG_SM_NAME_SMALL_CAMEL__] = camel_case_small(smmodel.statemachinename)
        dict_to_replace_lines[__TAG_SM_NAME__] = smmodel.statemachinename
        dict_to_replace_lines[__TAG_CLASS_NAME__] = smmodel.statemachinename
        dict_to_replace_lines[__TAG_PyIFGen_NAME__] = smmodel.pythoninterfacegeneratorfilename.replace('.py', '')  # hack : for tcpgen simple templates,
        if not dict_to_replace_lines[__TAG_PyIFGen_NAME__]:
            dict_to_replace_lines[__TAG_PyIFGen_NAME__] = self.vpp_filename
        dict_to_replace_lines[__TAG_NAMESPACE__] = smmodel.namespacename
        dict_to_replace_lines[__TAG_AUTHOR__] = self.author
        dict_to_replace_lines[__TAG_GROUP__] = self.group
        dict_to_replace_lines[__TAG_BRIEF__] = self.brief
        dict_to_replace_lines[__TAG_DECLSPEC_DLL_EXPORT__] = smmodel.declspecdllexport

        if self.language and self.events_interface:
            enums = ""
            for e in self.events_interface.Enums():
                enums += self.language.DeclareEnum(e, '\t')
            dict_to_replace_lines[__TAG_ENUMERATIONS__] = enums


        dict_to_replace_filenames = {}
        dict_to_replace_filenames["TEMPLATE_"] = smmodel.statemachinename

        return CGenerator.loadtemplates_firstfiltering(self,dict_to_replace_lines,dict_to_replace_filenames)

    def get_event_signature(self,name, with_defaults):
        if self.events_interface is None or self.language is None:
            return ""
        for s in self.events_interface.All():
            if s.Name == name:
                return self.language.ParameterString(self.language.GetFactoryCreateParams(s, self.events_interface, with_defaults))

        return ""

    def instantiate_event_struct_member(self, name, whitespace_cnt, is_ptr=True, instancename="data"):
        if self.events_interface is None or self.language is None:
            return ""
        for s in self.events_interface.All():
            if s.Name == name:
                guts = self.language.InstantiateStructMembers(s, self.events_interface, '', instancename, self.language.Accessor(is_ptr))
                result = ''
                cnt = 0
                for g in guts:
                    result = result + (whitespace_cnt*'    ' if cnt > 0 else '') + g + '\n'
                    cnt = cnt + 1
                return result.rsplit('\n', 1)[0]
        return ""

    def instantiate_event_aggregate_initializer(self, name):
        if self.events_interface is None or self.language is None:
            return ""
        for s in self.events_interface.All():
            if s.Name == name:
                return self.language.InstantiateStructWithAggregateInitializer(s, self.events_interface)
        return self.language.DefaultAggregateInitializer()

    def declare_event_struct_members(self, name, whitespace_cnt, attr_packed=False):
        if self.events_interface is None or self.language is None:
            return ""

        for s in self.events_interface.All():
            if s.Name == name:
                guts = self.language.DeclareStructMembers(s, self.events_interface, '', attr_packed)
                result = ''
                cnt = 0
                for g in guts:
                    result = result + ((whitespace_cnt+1)*'    ' if cnt > 0 else '    ') + g + '\n'
                    cnt = cnt + 1
                return result.rsplit('\n', 1)[0]
        return ""

    def innerexpand_secondfiltering(self, snippet_to_expand, alllinesexpanded, items):
        alpha = reset_alphabet()
        cnt = 0
        for name in items:
            for line in snippet_to_expand:
                # If there is no tag ... don't waste time
                if not hasTag(line):
                    if line.isspace():
                        continue
                    alllinesexpanded.append(line)
                    continue
                newline = line.replace(__TAG_STATENAME_SMALL_CAMEL__, camel_case_small(name))
                newline = newline.replace(__TAG_STATENAME__, name)
                newline = newline.replace(__TAG_EVENTNAME_SMALL_CAMEL__, camel_case_small(name))
                newline = newline.replace(__TAG_EVENTNAME__, name)
                newline = newline.replace(__TAG_ACTIONNAME__, name)
                newline = newline.replace(__TAG_ACTIONNAME_SMALL_CAMEL__, camel_case_small(name))
                newline = newline.replace(__TAG_GUARDNAME__, name)
                newline = newline.replace(__TAG_GUARDNAME_SMALL_CAMEL__, camel_case_small(name))
                newline = newline.replace(__TAG_ABC__, alpha)
                newline = newline.replace(__TAG_123__, str(cnt))
                tabcnt = newline.count('    ')
                if hasSpecificTag(newline,__TAG_SIGNATURE__):
                    has_signature_defaults = __TAG_SIGNATURE_DEF__ in newline
                    if hasDefault(newline):
                        # Has user params
                        user_params = extractDefaultAndTagNamed(newline, cleanTag(__TAG_SIGNATURE__))
                        signature = self.get_event_signature(name, False) + ", " + user_params[1]
                        signature = signature.strip(',').strip(' ')
                        newline = newline.replace(user_params[0], signature)
                    else:
                        newline = newline.replace(__TAG_SIGNATURE_DEF__ if has_signature_defaults else __TAG_SIGNATURE__, self.get_event_signature(name, has_signature_defaults))
                    # check for brackets...remove any spurious ',' and ' '
                    newline = re.sub("\([^)]*\)", lambda x:x.group(0).replace(' , )',')').replace(', )',')').replace(',)',')').replace('( , ','(').replace('( ,','(').replace('(,','('), newline)
                # __TAG_MEMBERINST__ -> PTR
                if hasSpecificTag(newline,__TAG_MEMBERINST__) and hasDefault(newline):
                    line_member = extractDefaultAndTag(newline)
                    newline = newline.replace(line_member[0],self.instantiate_event_struct_member(name, tabcnt, True, line_member[1]))
                else:
                    newline = newline.replace(__TAG_MEMBERINST__, self.instantiate_event_struct_member(name, tabcnt, True))        # PTR
                # __TAG_LITE_MEMBERINST__ -> NO PTR
                if hasSpecificTag(newline,__TAG_LITE_MEMBERINST__) and hasDefault(newline):
                    line_member = extractDefaultAndTag(newline)
                    newline = newline.replace(line_member[0],self.instantiate_event_struct_member(name, tabcnt, False, line_member[1]))
                else:
                    newline = newline.replace(__TAG_LITE_MEMBERINST__, self.instantiate_event_struct_member(name, tabcnt, False))  # NO PTR
                newline = newline.replace(__TAG_MEMBERDECL__, self.declare_event_struct_members(name, tabcnt))
                # Documentation
                if hasSpecificTag(newline, __TAG_DOCUMENTATION__):
                    ws = getWhitespace(newline)
                    docs = self.events_interface[name].documentation.rstrip("\n").split("\n")
                    for d in docs:
                        alllinesexpanded.append(ws + d + "\n")
                    continue
                # Aggregate initialization
                if hasSpecificTag(newline, __TAG_AGGREGATE_INIT__):
                    newline = newline.replace(__TAG_AGGREGATE_INIT__, self.instantiate_event_aggregate_initializer(name))
                # Member attributes ... all structs (msgs, protocols, structs)
                if hasSpecificTag(newline, __TAG_ATTRIBUTE_TYPE__) or hasSpecificTag(newline, __TAG_ATTRIBUTE_NAME__):
                    struct_or_msg = self.events_interface[name]
                    members = struct_or_msg.Decompose(False)
                    for mem in members:
                        membername = mem[1]
                        membertype = mem[0]
                        alllinesexpanded.append(newline.replace(__TAG_ATTRIBUTE_TYPE__, membertype).replace(__TAG_ATTRIBUTE_NAME__, membername))
                    continue
                # Python attributes ... someone piggybacks the nice interface
                if hasSpecificTag(newline, __TAG_PYTHON_ATTR__) and hasDefault(newline):
                    [tag, attr, useifnotexist] = extractTagAndAandB(newline)
                    if hasattr(self.events_interface[name], attr):
                        newline = newline.replace(tag, str(getattr(self.events_interface[name], attr)))
                    elif useifnotexist:
                        newline = newline.replace(tag, useifnotexist)
                    else:
                        continue
                if newline.isspace():
                    continue
                alllinesexpanded.append(newline)
            cnt = cnt + 1
            alpha = get_next_alphabet()

    ### CONSOLODATE ... this is essentially a copy-paste of the above ...
    def innerexpand_secondfiltering_PROTO(self, snippet_to_expand, alllinesexpanded, items):
        alpha = reset_alphabet()
        cnt = 0
        for name in items:
            for line in snippet_to_expand:
                # If there is no tag ... don't waste time
                if not hasTag(line):
                    if line.isspace():
                        continue
                    alllinesexpanded.append(line)
                    continue

                newline = line.replace(__TAG_STRUCTNAME_SMALL_CAMEL__, camel_case_small(name))
                newline = newline.replace(__TAG_STRUCTNAME__, name)
                newline = newline.replace(__TAG_MSGNAME_SMALL_CAMEL__, camel_case_small(name))
                newline = newline.replace(__TAG_MSGNAME__, name)
                newline = newline.replace(__TAG_PROTOMSGNAME__, name)
                newline = newline.replace(__TAG_PROTOMSGNAME_SMALL_CAMEL__, camel_case_small(name))
                newline = newline.replace(__TAG_ABC__, alpha)
                newline = newline.replace(__TAG_123__, str(cnt))
                tabcnt = newline.count('    ')
                if hasSpecificTag(newline, __TAG_SIGNATURE__):
                    has_signature_defaults = __TAG_SIGNATURE_DEF__ in newline
                    if hasDefault(newline):
                        # Has user params
                        user_params = extractDefaultAndTagNamed(newline, cleanTag(__TAG_SIGNATURE__))
                        signature = self.get_event_signature(name, False) + ", " + user_params[1]
                        signature = signature.strip(',').strip(' ')
                        newline = newline.replace(user_params[0], signature)
                    else:
                        newline = newline.replace(__TAG_SIGNATURE_DEF__ if has_signature_defaults else __TAG_SIGNATURE__, self.get_event_signature(name, has_signature_defaults))
                    # check for brackets...remove any spurious ',' and ' '
                    newline = re.sub("\([^)]*\)", lambda x: x.group(0).replace(' , )', ')').replace(', )', ')').replace(',)', ')').replace('( , ', '(').replace('( ,', '(').replace('(,', '('), newline)
                # __TAG_MEMBERINST__ -> PTR
                if hasSpecificTag(newline, __TAG_MEMBERINST__) and hasDefault(newline):
                    line_member = extractDefaultAndTag(newline)
                    newline = newline.replace(line_member[0], self.instantiate_event_struct_member(name, tabcnt, True, line_member[1]))
                else:
                    newline = newline.replace(__TAG_MEMBERINST__, self.instantiate_event_struct_member(name, tabcnt, True))  # PTR
                # __TAG_LITE_MEMBERINST__ -> NO PTR
                if hasSpecificTag(newline, __TAG_LITE_MEMBERINST__) and hasDefault(newline):
                    line_member = extractDefaultAndTag(newline)
                    newline = newline.replace(line_member[0], self.instantiate_event_struct_member(name, tabcnt, False, line_member[1]))
                else:
                    newline = newline.replace(__TAG_LITE_MEMBERINST__, self.instantiate_event_struct_member(name, tabcnt, False))  # NO PTR
                # Documentation
                if hasSpecificTag(newline, __TAG_DOCUMENTATION__):
                    ws = getWhitespace(newline)
                    docs = self.events_interface[name].documentation.rstrip("\n").split("\n")
                    for d in docs:
                        alllinesexpanded.append(ws + d + "\n")
                    continue
                # Aggregate initialization
                if hasSpecificTag(newline, __TAG_AGGREGATE_INIT__):
                    newline = newline.replace(__TAG_AGGREGATE_INIT__, self.instantiate_event_aggregate_initializer(name))
                if hasSpecificTag(newline, __TAG_MSGID__):
                    newline = newline.replace(__TAG_MSGID__, str(self.events_interface[name].MessageTypeID))
                newline = newline.replace(__TAG_MEMBERDECL__, self.declare_event_struct_members(name, tabcnt, True))
                # Member attributes ... all structs (msgs, protocols, structs)
                if hasSpecificTag(newline, __TAG_ATTRIBUTE_TYPE__) or hasSpecificTag(newline, __TAG_ATTRIBUTE_NAME__):
                    struct_or_msg = self.events_interface[name]
                    members = struct_or_msg.Decompose(False)
                    for mem in members:
                        membername = mem[1]
                        membertype = mem[0]
                        alllinesexpanded.append(newline.replace(__TAG_ATTRIBUTE_TYPE__,membertype).replace(__TAG_ATTRIBUTE_NAME__, membername))
                    continue
                # Payload attributes only
                if hasSpecificTag(newline, __TAG_PAYLOAD_TYPE__) or hasSpecificTag(newline, __TAG_PAYLOAD_NAME__):
                    struct_or_msg = self.events_interface[name]
                    structmembers = struct_or_msg.Decompose(False)
                    for mem in structmembers:
                        membername = mem[1]
                        membertype = mem[0]
                        #isStruct = struct_or_msg.IsStruct(membername)
                        isProtocol = struct_or_msg.IsProtocolStruct(membername)
                        if not isProtocol:
                            alllinesexpanded.append(newline.replace(__TAG_PAYLOAD_TYPE__, membertype).replace(__TAG_PAYLOAD_NAME__, membername))
                    continue
                # Python attributes ... someone piggybacks the nice interface
                if hasSpecificTag(newline,__TAG_PYTHON_ATTR__) and hasDefault(newline):
                    [tag, attr, useifnotexist] = extractTagAndAandB(newline)
                    if hasattr(self.events_interface[name], attr):
                        newline = newline.replace(tag, str(getattr(self.events_interface[name], attr)))
                    elif useifnotexist:
                        newline = newline.replace(tag, useifnotexist)
                    else:
                        continue
                if newline.isspace():
                    continue
                alllinesexpanded.append(newline)
            cnt = cnt + 1
            alpha = get_next_alphabet()
    ###

    def innerexpand_actionsignatures(self, snippet_to_expand, alllinesexpanded, states):
        alpha = reset_alphabet()
        cnt = 0
        for key, (actionname, eventname) in states.items():
            if eventname == "" or eventname.lower() == 'none':
                eventname = "NONE"
            elif eventname.lower() == 'any':
                eventname = "ANY"
            for line in snippet_to_expand:
                alllinesexpanded.append(line
                            .replace(__TAG_ACTIONNAME_SMALL_CAMEL__, camel_case_small(actionname))
                            .replace(__TAG_ACTIONNAME__, actionname)
                            .replace(__TAG_EVENTNAME_SMALL_CAMEL__, camel_case_small(eventname))
                            .replace(__TAG_EVENTNAME__, eventname)
                            .replace(__TAG_ABC__, alpha)
                            .replace(__TAG_123__, str(cnt)))
            cnt = cnt + 1
            alpha = get_next_alphabet()

    def innerexpand_msm(self, output, whitespace, smmodel):
        len_tt = len(smmodel.transition_table)
        tt_out = whitespace + "// " + len("msmf::Row < ") * ' ' + even_space("Start") + even_space("Event") + even_space("Next") + even_space("Action") + even_space("Guard") + '\n'
        for i, ttline in enumerate(smmodel.transition_table):
            tt_out += whitespace + 'msmf::Row < '
            tt_out += even_space(self.transitiontable_replace_NONE(ttline[smmodel.START_STATE])) + ','
            tt_out += even_space(self.transitiontable_replace_NONE(ttline[smmodel.EVENT])) + ','
            tt_out += even_space(self.transitiontable_replace_NONE(ttline[smmodel.NEXT_STATE])) + ','
            tt_out += even_space(self.transitiontable_replace_NONE(ttline[smmodel.ACTION])) + ','
            tt_out += even_space(self.transitiontable_replace_NONE(ttline[smmodel.GUARD])) + '>    '
            if i != len_tt - 1:
                tt_out += ","
            tt_out += "    // " + str(i) + '\n'
            output.append(tt_out)
            tt_out = ""

    def innerexpand_msmlite(self, output, whitespace, smmodel):
        tt_out = whitespace + "// " + even_space("Start + ") + even_space("Event") + even_space("[ Guard ] ") + even_space("/ Action") + even_space(" = Next") + '\n'
        startStateHasEntryExit = {}
        for i, ttline in enumerate(smmodel.transition_table):
            if i == 0:  # initial state
                tt_out += whitespace + " *"
            else:
                tt_out += whitespace + ", "
            tt_out += even_space(self.transitiontable_replace_NONE(ttline[smmodel.START_STATE])) + '+'
            tt_out += even_space('event<' + self.transitiontable_replace_NONE(ttline[smmodel.EVENT]) + ">") + ' '
            tt_out += even_space('[' + self.transitiontableLITE_guard_replace_NONE('__' + ttline[smmodel.GUARD]) + ']') + ' / '
            tt_out += even_space(self.transitiontableLITE_action_replace_NONE('__' + ttline[smmodel.ACTION]))
            if ttline[smmodel.NEXT_STATE].lower() != 'none':  # to not get transitions into/outof state on actions that dont change the state...
                tt_out += ' = ' + even_space(self.transitiontableLITE_nextstate_replace_NONE(ttline[smmodel.NEXT_STATE], ttline[smmodel.START_STATE]))
            tt_out = tt_out.rstrip()
            tt_out += '\n'
            output.append(tt_out)
            tt_out = ""
            # State entry/exit, once only
            if not (ttline[smmodel.START_STATE] in startStateHasEntryExit):
                startStateHasEntryExit[ttline[smmodel.START_STATE]] = True
                tt_out += whitespace + ", " + ttline[smmodel.START_STATE] + " + msm::on_entry / __" + ttline[smmodel.START_STATE] + 'OnEntry\n'
                tt_out += whitespace + ", " + ttline[smmodel.START_STATE] + " + msm::on_exit / __" + ttline[smmodel.START_STATE] + 'OnExit'
                tt_out = tt_out.rstrip()
                tt_out += '\n'
                output.append(tt_out)
                tt_out = ""

    def innerexpand_plant(self, output, whitespace, smmodel):
        lines = TTToDot(smmodel.transition_table)
        for l in lines:
            output.append(whitespace + l + "\n")

    def innerexpand_sml(self, output, whitespace, smmodel, sml_entry_exit):
        tt_out = whitespace + "// " + even_space("Start", smmodel.maxlenSTART_STATE + 8) + even_space("+Event", smmodel.maxlenEVENT + 10) + even_space("[ Guard ]", smmodel.maxlenGUARD + 6) + even_space("/ Action", smmodel.maxlenACTION + 4) + even_space(" = Next", 0) + '\n'
        startStateHasEntryExit = {}
        for i, ttline in enumerate(smmodel.transition_table):
            if i == 0:  # initial state
                tt_out += whitespace + " *"
            else:
                tt_out += whitespace + ", "
            tt_out += even_space('state<' + self.transitiontable_replace_NONE(ttline[smmodel.START_STATE]) + '>', smmodel.maxlenSTART_STATE + 9) + '+'
            tt_out += even_space('event<' + self.transitiontable_replace_NONE(ttline[smmodel.EVENT]) + '>', smmodel.maxlenEVENT + 9) + ' '
            tt_out += even_space('[' + self.transitiontableLITE_guard_replace_NONE(camel_case_small(ttline[smmodel.GUARD])) + ']', smmodel.maxlenGUARD + 4) + ' / '
            tt_out += even_space(self.transitiontableLITE_action_replace_NONE(camel_case_small(ttline[smmodel.ACTION])), smmodel.maxlenACTION + 2)
            if ttline[smmodel.NEXT_STATE].lower() != 'none':  # to not get transitions into/outof state on actions that dont change the state...
                tt_out += ' = ' + even_space('state<' + self.transitiontableLITE_nextstate_replace_NONE(ttline[smmodel.NEXT_STATE], ttline[smmodel.START_STATE]) + '>', 0)
            tt_out = tt_out.rstrip()
            tt_out += '\n'
            output.append(tt_out)
            tt_out = ""
            # State entry/exit, once only
            if not (ttline[smmodel.START_STATE] in startStateHasEntryExit) and sml_entry_exit:
                startStateHasEntryExit[ttline[smmodel.START_STATE]] = True
                tt_out += whitespace + ", state<" + ttline[smmodel.START_STATE] + "> + boost::sml::on_entry<_> / " + camel_case_small(ttline[smmodel.START_STATE]) + 'OnEntry\n'
                tt_out += whitespace + ", state<" + ttline[smmodel.START_STATE] + "> + boost::sml::on_exit<_> / " + camel_case_small(ttline[smmodel.START_STATE]) + 'OnExit'
                tt_out = tt_out.rstrip()
                tt_out += '\n'
                output.append(tt_out)
                tt_out = ""

    def filterInitialState(self, all_lines, smmodel):
        output = []
        for l in all_lines:
            output.append(l.replace(__TAG_INIT_STATE__, smmodel.getfirststate()).replace(__TAG_INIT_STATE_SMALL_CAMEL__, camel_case_small(smmodel.getfirststate())))
        return output


    def filterStateName(self, lines, stateName):
        result = []
        for l in lines:
            result.append(l.replace(__TAG_STATENAME__, stateName).replace(__TAG_STATENAME_SMALL_CAMEL__, camel_case_small(stateName)))
        return result


    def filterEventName(self, lines, eventName):
        result = []
        for l in lines:
            result.append(l.replace(__TAG_EVENTNAME__, eventName).replace(__TAG_EVENTNAME_SMALL_CAMEL__, camel_case_small(eventName)))
        return result


    def innerexpand_transitionsperstate(self, snippet_to_expand, all_lines_expanded, transitionperstate):

        def __expansion(to_expand, output, object, state, transition_dict):
            for ev, transitionList in transition_dict.items():
                # guard/action/next state repeats
                object.innerexpand_transitionsperguard(to_expand, output, ev, state, transitionList)

        for state, transition_dict in transitionperstate.items():
            all_lines_snippet = self.filterStateName(snippet_to_expand, state)
            all_lines_snippet = PairExpander(__TAG_PET_BEGIN__, __TAG_PET_END__).Expand(all_lines_snippet, __expansion, self, state, transition_dict)
            all_lines_expanded.extend(all_lines_snippet)


    def innerexpand_transitionsperguard(self, snippet_to_expand, all_lines_expanded, eventName, stateName, transitionList):

        def __expansion(to_expand, output, transitionList):
            for transitionDict in transitionList:
                for l in to_expand:
                    for k, v in transitionDict.items():
                        # for those that are present but who have alternate text when not present.
                        if hasSpecificTag(l, k):
                            l = removeDefault(l)
                        l = l.replace(k, v)
                    # l = l.replace(__TAG_EVENTNAME__, eventName)
                    # l = l.replace(__TAG_EVENTNAME_SMALL_CAMEL__, camel_case_small(eventName))
                    # If there is no guard, or no next state, or no action, just remove it (or replace it with the alternative text). Leave no hanging code.
                    if hasSpecificTag(l, __TAG_GUARDNAME_SMALL_CAMEL__) or hasSpecificTag(l, __TAG_GUARDNAME__) or hasSpecificTag(l, __TAG_NEXTSTATENAME__) or hasSpecificTag(l, __TAG_ACTIONNAME__) or hasSpecificTag(l, __TAG_STATENAME_IF_NEXTSTATE__) or hasSpecificTag(l, __TAG_STATENAME_IF_NEXTSTATE_SMALL_CAMEL__):
                        line_member = extractDefaultAndTag(l)
                        if line_member[1]:  # alternative text is embedded in the tag.
                            whitespace = len(l) - len(l.lstrip())
                            output.append(whitespace * ' ' + line_member[1] + '\n')
                    elif l.find(__TAG_GUARDNAME_SMALL_CAMEL__) == -1 and l.find(__TAG_GUARDNAME__) == -1 and l.find(__TAG_NEXTSTATENAME__) == -1 and l.find(__TAG_STATENAME_IF_NEXTSTATE__) == -1 and l.find(__TAG_STATENAME_IF_NEXTSTATE_SMALL_CAMEL__) == -1:
                        output.append(l)

        all_lines_snippet = self.filterEventName(snippet_to_expand, eventName)
        all_lines_snippet = PairExpander(__TAG_PGT_BEGIN__, __TAG_PGT_END__).Expand(all_lines_snippet, __expansion, transitionList)
        all_lines_expanded.extend(all_lines_snippet)

    def transitiontable_replace_NONE(self, val):
        if val == "" or val.lower() == 'none':
            val = "msmf::none"
        return val

    def transitiontableLITE_guard_replace_NONE(self, val):
        tmp_val = val.replace('__', '')
        if tmp_val == "" or tmp_val.lower() == 'none':
            val = "gnone"
        return val

    def transitiontableLITE_action_replace_NONE(self, val):
        tmp_val = val.replace('__', '')
        if tmp_val == "" or tmp_val.lower() == 'none' or tmp_val.lower().find('::none<') > -1:
            val = "none"
        return val

    ''' This SM doesnt seem to allow 'none' transitions -> make it transition to the source state'''
    def transitiontableLITE_nextstate_replace_NONE(self, val, source_state):
        tmp_val = val.replace('__', '')
        tmp_val = tmp_val.replace('msmf::', '')
        if tmp_val == "" or tmp_val.lower() == 'none':
            val = source_state
        return val


    def expand_secondfiltering(self, smmodel, cmmodel):
        for file in cmmodel.filenames_to_lines:
            all_lines_expanded = self.filterInitialState(cmmodel.filenames_to_lines[file], smmodel)
            all_lines_expanded = SingleExpander(__TAG_TTT_PLANT_UML__).Expand(all_lines_expanded, self.innerexpand_plant, smmodel)
            all_lines_expanded = SingleExpander(__TAG_TTT_BOOST_MSM__).Expand(all_lines_expanded, self.innerexpand_msm, smmodel)
            all_lines_expanded = SingleExpander(__TAG_TTT_BOOST_MSMLITE__).Expand(all_lines_expanded, self.innerexpand_msmlite, smmodel)
            all_lines_expanded = SingleExpander(__TAG_TTT_BOOST_SML_ENTRYEXIT__).Expand(all_lines_expanded, self.innerexpand_sml, smmodel, True)
            all_lines_expanded = SingleExpander(__TAG_TTT_BOOST_SML__).Expand(all_lines_expanded, self.innerexpand_sml, smmodel, False)
            all_lines_expanded = PairExpander(__TAG_PS_BEGIN__, __TAG_PS_END__).Expand(all_lines_expanded, self.innerexpand_secondfiltering, smmodel.states)
            all_lines_expanded = PairExpander(__TAG_PE_BEGIN__, __TAG_PE_END__).Expand(all_lines_expanded, self.innerexpand_secondfiltering, smmodel.events)
            all_lines_expanded = PairExpander(__TAG_PA_BEGIN__, __TAG_PA_END__).Expand(all_lines_expanded, self.innerexpand_secondfiltering, smmodel.actions)
            all_lines_expanded = PairExpander(__TAG_PASIG_BEGIN__, __TAG_PASIG_END__).Expand(all_lines_expanded, self.innerexpand_actionsignatures, smmodel.actionsignatures)
            all_lines_expanded = PairExpander(__TAG_PST_BEGIN__, __TAG_PST_END__).Expand(all_lines_expanded, self.innerexpand_transitionsperstate, smmodel.transitionsperstate)
            all_lines_expanded = PairExpander(__TAG_PG_BEGIN__, __TAG_PG_END__).Expand(all_lines_expanded, self.innerexpand_secondfiltering, smmodel.guards)

            ### CONSOLODATE
            if self.events_interface:
                all_lines_expanded = PairExpander(__TAG_STRUCT_BEGIN__, __TAG_STRUCT_END__).Expand(all_lines_expanded, self.innerexpand_secondfiltering_PROTO, self.events_interface.StructNames())
                all_lines_expanded = PairExpander(__TAG_PROTOMSG_BEGIN__, __TAG_PROTOMSG_END__).Expand(all_lines_expanded, self.innerexpand_secondfiltering_PROTO, self.events_interface.ProtocolStructNames())
                all_lines_expanded = PairExpander(__TAG_MSG_BEGIN__, __TAG_MSG_END__).Expand(all_lines_expanded, self.innerexpand_secondfiltering_PROTO, self.events_interface.MessageNames())
            ###
            cmmodel.filenames_to_lines[file] = all_lines_expanded

    ''' Used for State Machine Generation
    '''
    def Generate(self, transitiontable, namespacenname, statemachinename, dclspc="", copyotherfiles = True) -> list:

        print("*************************************")
        print("******* SMGen ***********************")
        print("*************************************")
        print(" Output Dir   : " + self.output_gen_file_dir)
        print(" State Machine: " + statemachinename)
        print(" Executing in : " + os.path.realpath(__file__))
        print("*************************************")

        sm = CTransitionTableModel(transitiontable, namespacenname, statemachinename, dclspc)

        for e in self.events_interface.Structs(): # only necessary for stateless (nonTT) events.
            if not e.Name in sm.events:
                sm.events.append(e.Name)

        cm = self.loadtemplates_firstfiltering(sm)
        self.expand_secondfiltering(sm, cm)

        # user tags.
        if self.events_interface != None:
            self.do_user_tags(cm, self.events_interface.UserTags())

        # For processing
        self.do_for(cm)

        # Preserve user code.
        self.preserve_usercode_in_files(cm)
        '''
        # Round-trip Code Preservation. Will load the code to preserve upon creation (if the output dir is not-empty/the same as the one in the compile path).
        preservation = Preservative(self.output_gen_file_dir)
        preservation.Emplace(cm.filenames_to_lines)
        '''
        # Write output to file.
        res = self.createoutput(cm.filenames_to_lines)

        # Copy non-autogenerated required files to output.
        if isinstance(self.language, LanguageCPP) and copyotherfiles:

            # Files...
            files_to_copy = []
            files_to_copy.append("allocator.h")
            files_to_copy.append("allocator.cpp")
            files_to_copy.append("basetypes.h")
            files_to_copy.append("CMakeLists.txt")
            files_to_copy.append("Fault.h")
            files_to_copy.append("Fault.cpp")
            files_to_copy.append("stl_allocator.h")
            files_to_copy.append("thread_FreeRTOS.h")
            files_to_copy.append("thread_FreeRTOS.cpp")
            files_to_copy.append("threaded_dispatcher.h")
            files_to_copy.append("threaded_dispatcher_FreeRTOS.h")
            files_to_copy.append("threadsafe_queue.h")
            files_to_copy.append("threadsafe_queue_FreeRTOS.h")
            files_to_copy.append("waitcondition.h")
            files_to_copy.append("waitcondition.cpp")
            files_to_copy.append("xallocator.h")
            files_to_copy.append("xallocator.cpp")
            files_to_copy.append("xlist.h")
            files_to_copy.append("xmap.h")
            files_to_copy.append("xqueue.h")
            files_to_copy.append("xset.h")
            files_to_copy.append("xsstream.h")
            files_to_copy.append("xstring.h")

            allplatformsfrom = os.path.join(os.path.abspath(os.path.dirname(__file__)), os.path.join("allplatforms", "CPP"))
            allplatformsto = os.path.join(os.path.abspath(self.output_gen_file_dir), "allplatforms")

            FileCopyUtil(allplatformsfrom, allplatformsto, files_to_copy)

            # Boost SML ...
            smlfrom = os.path.join(allplatformsfrom,	os.path.join("sml", os.path.join("include","boost")))
            smlto = os.path.join(allplatformsto, "boost")
            smlfiles_to_copy = []
            smlfiles_to_copy.append("sml.hpp")
            FileCopyUtil(smlfrom, smlto, smlfiles_to_copy)

            # Tests...
            testfiles_to_copy = []
            testfiles_to_copy.append("CMakeLists.txt")
            testfiles_to_copy.append("Test.ThreadingConcepts.cpp")
            testfiles_to_copy.append("test_main.cpp")

            tests_allplatformsfrom = os.path.join(allplatformsfrom, "testsuite")
            tests_allplatformsto = os.path.join(allplatformsto, "testsuite")

            FileCopyUtil(tests_allplatformsfrom, tests_allplatformsto, testfiles_to_copy)

            # Micro Unit Test Framework
            microunit_files_to_copy = []
            microunit_files_to_copy.append("minunit.h")
            microunit_files_to_copy.append("minunit.cpp")

            microunit_allplatformsfrom = os.path.join(tests_allplatformsfrom, "minunit")
            microunit_allplatformsto = os.path.join(tests_allplatformsto, "minunit")

            FileCopyUtil(microunit_allplatformsfrom, microunit_allplatformsto, microunit_files_to_copy)

        return res


    ''' Used for Protocol Generation
    '''
    def GenerateProtocol(self, pythoninterfacegeneratorfilename, namespacenname, classname, dclspc="", preserve_dir="") -> list:
        sm = CTransitionTableModel([], namespacenname, classname, dclspc)
        sm.pythoninterfacegeneratorfilename = pythoninterfacegeneratorfilename
        cm = self.loadtemplates_firstfiltering(sm)
        self.expand_secondfiltering(sm, cm)

        # Round-trip Code Preservation. Will load the code to preserve upon creation (if the output dir is not-empty/the same as the one in the compile path).
        # TCP gen might have a different output directory (typically COG will put files into an intermediate dir, and them copy them elsewhere
        preservation = None
        if preserve_dir == "":
            preservation = Preservative(self.output_gen_file_dir)
        else:
            preservation = Preservative(preserve_dir)

        preservation.Emplace(cm.filenames_to_lines)

        return self.createoutput(cm.filenames_to_lines)

